# Copyright (c) 2013 Marion Zepf
# Copyright (c) 2014 Walter Bender

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import ast
from gettext import gettext as _
from math import sqrt
from random import uniform
import traceback
import inspect

# from ast_pprint import * # only used for debugging, safe to comment out


from .tablock import Media

from .tacanvas import TurtleGraphics
from .taconstants import (Color, CONSTANTS, ColorObj, Vector)
from .talogo import (LogoCode, logoerror, NegativeRootError)
from .taturtle import (Turtle, Turtles)
from TurtleArt.tatype import (TYPE_CHAR, TYPE_INT, TYPE_FLOAT, TYPE_OBJECT,
                              TYPE_MEDIA, TYPE_COLOR, BOX_AST, ACTION_AST,
                              TYPE_VECTOR,
                              Type, TypeDisjunction, TATypeError, get_type,
                              TypedSubscript, TypedName, is_bound_method,
                              is_instancemethod, is_staticmethod,
                              identity, get_converter, convert, get_call_ast)
from .tautils import debug_output
from .tawindow import (TurtleArtWindow, global_objects, plugins_in_use)
from .util import ast_extensions


class PyExportError(BaseException):

    """ Error that is raised when something goes wrong while converting the
    blocks to python code """

    def __init__(self, message, block=None):
        """ message -- the error message
        block -- the block where the error occurred """
        self.message = message
        self.block = block

    def __str__(self):
        if self.block is not None:
            return _("error in highlighted block") + ": " + str(self.message)
        else:
            return _("error") + ": " + str(self.message)


class Primitive(object):

    """ Something that can be called when the block code is executed in TA,
    but that can also be transformed into a Python AST."""

    _DEBUG = False

    STANDARD_OPERATORS = {'plus': (ast.UAdd, ast.Add),
                          'minus': (ast.USub, ast.Sub),
                          'multiply': ast.Mult,
                          'divide': ast.Div,
                          'modulo': ast.Mod,
                          'power': ast.Pow,
                          'and_': ast.And,
                          'or_': ast.Or,
                          'not_': ast.Not,
                          'equals': ast.Eq,
                          'less': ast.Lt,
                          'greater': ast.Gt}

    def __init__(self, func, return_type=TYPE_OBJECT, arg_descs=None,
                 kwarg_descs=None, call_afterwards=None, export_me=True):
        """ return_type -- the type (from the type hierarchy) that this
            Primitive will return
        arg_descs, kwarg_descs -- a list of argument descriptions and
            a dictionary of keyword argument descriptions. An argument
            description can be either an ArgSlot or a ConstantArg.
        call_afterwards -- Code to call after this Primitive has been called
            (e.g., for updating labels in LogoCode) (not used for creating
            AST)
        export_me -- True iff this Primitive should be exported to Python
            code (the default case) """
        self.func = func
        self.return_type = return_type

        if arg_descs is None:
            self.arg_descs = []
        else:
            self.arg_descs = arg_descs

        if kwarg_descs is None:
            self.kwarg_descs = {}
        else:
            self.kwarg_descs = kwarg_descs

        self.call_afterwards = call_afterwards
        self.export_me = export_me

    def copy(self):
        """ Return a Primitive object with the same attributes as this one.
        Shallow-copy the arg_descs and kwarg_descs attributes. """
        arg_descs_copy = self.arg_descs[:]
        if isinstance(self.arg_descs, ArgListDisjunction):
            arg_descs_copy = ArgListDisjunction(arg_descs_copy)
        return Primitive(self.func,
                         return_type=self.return_type,
                         arg_descs=arg_descs_copy,
                         kwarg_descs=self.kwarg_descs.copy(),
                         call_afterwards=self.call_afterwards,
                         export_me=self.export_me)

    def __repr__(self):
        return "Primitive(%s -> %s)" % (repr(self.func), str(self.return_type))

    @property
    def __name__(self):
        return self.func.__name__

    def get_name_for_export(self):
        """ Return the expression (as a string) that represents this Primitive
        in the exported Python code, e.g., 'turtle.forward'. """
        func_name = ""
        if self.wants_turtle():
            func_name = "turtle."
        elif self.wants_turtles():
            func_name = "turtles."
        elif self.wants_canvas():
            func_name = "canvas."
        elif self.wants_logocode():
            func_name = "logo."
        elif self.wants_heap():
            func_name = "logo.heap."
        elif self.wants_tawindow():
            func_name = "tw."
        else:
            results, plugin = self.wants_plugin()
            if results:
                for k in list(global_objects.keys()):
                    if k == plugin:
                        if k not in plugins_in_use:
                            plugins_in_use.append(k)
                        func_name = k.lower() + '.'
                        break

        # get the name of the function directly from the function itself
        func_name += self.func.__name__
        return func_name

    def are_slots_filled(self):
        """ Return True iff none of the arg_descs or kwarg_descs is an
        ArgSlot. """
        for arg_desc in self.arg_descs:
            if isinstance(arg_desc, ArgSlot):
                return False
        for key in self.kwarg_descs:
            if isinstance(self.kwarg_descs[key], ArgSlot):
                return False
        return True

    def fill_slots(self, arguments=None, keywords=None, convert_to_ast=False,
                   call_my_args=True):
        """ Return a copy of this Primitive whose ArgSlots are filled with
        the given arguments, turned into ConstantArgs. Call the arguments,
        apply their wrappers, and check their types as appropriate. """
        if arguments is None:
            arguments = []
        if keywords is None:
            keywords = {}

        new_prim = self.copy()

        if isinstance(new_prim.arg_descs, ArgListDisjunction):
            slot_list_alternatives = list(new_prim.arg_descs)
        else:
            slot_list_alternatives = [new_prim.arg_descs]

        # arguments
        error = None
        filler = None
        for slot_list in slot_list_alternatives:
            error = None
            new_slot_list = []
            filler_list = list(arguments[:])
            for slot in slot_list:
                if isinstance(slot, ArgSlot):
                    filler = filler_list.pop(0)
                    try:
                        const = slot.fill(filler,
                                          convert_to_ast=convert_to_ast,
                                          call_my_args=call_my_args)
                    except TATypeError as e:
                        error = e
                        if Primitive._DEBUG:
                            traceback.print_exc()
                        break
                    else:
                        new_slot_list.append(const)
                else:
                    new_slot_list.append(slot)
            if error is None:
                new_prim.arg_descs = new_slot_list
                break
        if error is not None:
            raise error

        # keyword arguments
        for key in keywords:
            kwarg_desc = new_prim.kwarg_descs[key]
            if isinstance(kwarg_desc, ArgSlot):
                const = kwarg_desc.fill(keywords[key],
                                        convert_to_ast=convert_to_ast,
                                        call_my_args=call_my_args)
                new_prim.kwarg_descs[key] = const

        return new_prim

    def get_values_of_filled_slots(self, exportable_only=False):
        """ Return the values of all filled argument slots as a list, and
        the values of all filled keyword argument slots as a dictionary.
        Ignore all un-filled (keyword) argument slots.
        exportable_only -- return only exportable values and convert values
            to ASTs instead of calling them """
        new_args = []
        for c_arg in self.arg_descs:
            if (isinstance(c_arg, ConstantArg) and
                (not exportable_only or
                 export_me(c_arg.value))):
                new_args.append(c_arg.get(convert_to_ast=exportable_only))
        new_kwargs = {}
        for key in self.kwarg_descs:
            if (isinstance(self.kwarg_descs[key], ConstantArg) and
                (not exportable_only or
                 export_me(self.kwarg_descs[key].value))):
                new_kwargs[key] = self.kwarg_descs[key].get(
                    convert_to_ast=exportable_only)
        return (new_args, new_kwargs)

    def allow_call_args(self, recursive=False):
        """ Set call_args attribute of all argument descriptions to True
        recursive -- recursively call allow_call_args on all constant args
            that are Primitives """
        for arg_desc in self.arg_descs:
            arg_desc.call_arg = True
            if (recursive and isinstance(arg_desc, ConstantArg) and
                    isinstance(arg_desc.value, Primitive)):
                arg_desc.value.allow_call_args(recursive=True)
        for kwarg_desc in self.kwarg_descs:
            kwarg_desc.call_arg = True
            if (recursive and isinstance(kwarg_desc, ConstantArg) and
                    isinstance(kwarg_desc.value, Primitive)):
                kwarg_desc.value.allow_call_args(recursive=True)

    def __call__(self, *runtime_args, **runtime_kwargs):
        """ Execute the function, passing it the arguments received at
        runtime. Also call the function in self.call_afterwards and pass it
        all runtime_args and runtime_kwargs.
        If the very first argument is a LogoCode instance, it is removed.
        The active turtle, the Turtles object, the canvas, the LogoCode
        object, or the TurtleArtWindow object will be prepended to the
        arguments (depending on what this Primitive wants). """

        # remove the first argument if it is a LogoCode instance
        if runtime_args and isinstance(runtime_args[0], LogoCode):
            runtime_args = runtime_args[1:]

        if Primitive._DEBUG:
            debug_output(repr(self))
            debug_output("  runtime_args: " + repr(runtime_args))
        # fill the ArgSlots with the runtime arguments
        new_prim = self.fill_slots(runtime_args, runtime_kwargs,
                                   convert_to_ast=False)
        if not new_prim.are_slots_filled():
            raise logoerror("#syntaxerror")
        if Primitive._DEBUG:
            debug_output("  new_prim.arg_descs: " + repr(new_prim.arg_descs))

        # extract the actual values from the (now constant) arguments
        (new_args, new_kwargs) = new_prim.get_values_of_filled_slots()
        if Primitive._DEBUG:
            debug_output("  new_args: " + repr(new_args))
            debug_output("end " + repr(self))

        # what does this primitive want as its first argument?
        first_arg = None
        if not is_bound_method(new_prim.func):
            if new_prim.wants_turtle():
                first_arg = global_objects["turtles"].get_active_turtle()
            elif new_prim.wants_turtles():
                first_arg = global_objects["turtles"]
            elif new_prim.wants_canvas():
                first_arg = global_objects["canvas"]
            elif new_prim.wants_logocode():
                first_arg = global_objects["logo"]
            elif new_prim.wants_heap():
                first_arg = global_objects["logo"].heap
            elif new_prim.wants_tawindow():
                first_arg = global_objects["window"]
            else:
                result, plugin = new_prim.wants_plugin()
                if result:
                    first_arg = plugin

        # execute the actual function
        if first_arg is None:
            return_value = new_prim.func(*new_args, **new_kwargs)
        else:
            return_value = new_prim.func(first_arg, *new_args, **new_kwargs)

        if new_prim.call_afterwards is not None:
            new_prim.call_afterwards(*new_args, **new_kwargs)

        return return_value

    def get_ast(self, *arg_asts, **kwarg_asts):
        """Transform this object into a Python AST. When serialized and
        executed, the AST will do exactly the same as calling this
        object."""

        if Primitive._DEBUG:
            debug_output(repr(self))
            debug_output("  arg_asts: " + repr(arg_asts))
        new_prim = self.fill_slots(arg_asts, kwarg_asts, convert_to_ast=True)
        if not new_prim.are_slots_filled():
            raise PyExportError("not enough arguments")
        if Primitive._DEBUG:
            debug_output("  new_prim.arg_descs: " + repr(new_prim.arg_descs))

        # extract the actual values from the (now constant) arguments
        (new_arg_asts, new_kwarg_asts) = new_prim.get_values_of_filled_slots(
            exportable_only=True)
        if Primitive._DEBUG:
            debug_output("  new_arg_asts: " + repr(new_arg_asts))
            debug_output("end " + repr(self))

        # SPECIAL HANDLING #

        # loops
        if self == LogoCode.prim_loop:
            controller = self._get_loop_controller()
            if controller == Primitive.controller_repeat:
                # 'repeat' loop
                num_repetitions = new_arg_asts[0]
                if num_repetitions.func.id == 'controller_repeat':
                    num_repetitions = num_repetitions.args[0]
                repeat_iter = get_call_ast("range", [num_repetitions])
                # TODO use new variable name in nested loops
                loop_ast = ast.For(target=ast.Name(id="i", ctx=ast.Store),
                                   iter=repeat_iter,
                                   body=new_arg_asts[1],
                                   orelse=[])
                return loop_ast
            else:
                if controller == Primitive.controller_forever:
                    condition_ast = ast.Name(id="True", ctx=ast.Load)
                elif controller == Primitive.controller_while:
                    condition_ast = new_arg_asts[0].args[0]
                elif controller == Primitive.controller_until:
                    pos_cond_ast = new_arg_asts[0].args[0]
                    condition_ast = ast.UnaryOp(op=ast.Not,
                                                operand=pos_cond_ast)
                else:
                    raise PyExportError("unknown loop controller: " +
                                        repr(controller))
                loop_ast = ast.While(test=condition_ast,
                                     body=new_arg_asts[1],
                                     orelse=[])
                # Until always executes its body once.
                if controller == Primitive.controller_until:
                    loop_list = []
                    for arg_ast in new_arg_asts[1]:
                        loop_list.append(arg_ast)
                    loop_list.append(loop_ast)
                    return loop_list
                else:
                    return loop_ast

        # conditionals
        elif self in (LogoCode.prim_if, LogoCode.prim_ifelse):
            test = new_arg_asts[0]
            body = new_arg_asts[1]
            if len(new_arg_asts) > 2:
                orelse = new_arg_asts[2]
            else:
                orelse = []
            if_ast = ast.If(test=test, body=body, orelse=orelse)
            return if_ast

        # boxes
        elif self == LogoCode.prim_set_box:
            target_ast = ast.Subscript(value=BOX_AST,
                                       slice=ast.Index(value=new_arg_asts[0]),
                                       ctx=ast.Store)
            return ast.Assign(targets=[target_ast], value=new_arg_asts[1])
        elif self == LogoCode.prim_get_box:
            return ast.Subscript(value=BOX_AST,
                                 slice=ast.Index(value=new_arg_asts[0]),
                                 ctx=ast.Load)

        # action stacks
        elif self == LogoCode.prim_define_stack:
            return
        elif self == LogoCode.prim_invoke_stack:
            stack_func = ast.Subscript(
                value=ACTION_AST,
                slice=ast.Index(value=new_arg_asts[0]), ctx=ast.Load)
            call_ast = get_call_ast('logo.icall', [stack_func])
            return [call_ast, ast_yield_true()]
        elif self == LogoCode.prim_invoke_return_stack:
            # FIXME: Need to return value
            stack_func = ast.Subscript(
                value=ACTION_AST,
                slice=ast.Index(value=new_arg_asts[0]), ctx=ast.Load)
            call_ast = get_call_ast('logo.icall', [stack_func])
            return [call_ast, ast_yield_true()]

        # stop stack
        elif self == LogoCode.prim_stop_stack:
            return ast.Return()

        # sleep/ wait
        elif self == LogoCode.prim_wait:
            return [get_call_ast('sleep', new_arg_asts), ast_yield_true()]

        # standard operators
        elif self.func.__name__ in Primitive.STANDARD_OPERATORS:
            op = Primitive.STANDARD_OPERATORS[self.func.__name__]
            # 'divide': prevent unwanted integer division
            if self == Primitive.divide:
                def _is_float(x):
                    return get_type(x)[0] == TYPE_FLOAT
                if (not _is_float(new_arg_asts[0]) and
                        not _is_float(new_arg_asts[1])):
                    new_arg_asts[0] = get_call_ast('float', [new_arg_asts[0]],
                                                   return_type=TYPE_FLOAT)
            if len(new_arg_asts) == 1:
                if isinstance(op, tuple):
                    op = op[0]
                return ast.UnaryOp(op=op, operand=new_arg_asts[0])
            elif len(new_arg_asts) == 2:
                if isinstance(op, tuple):
                    op = op[1]
                (left, right) = new_arg_asts
                if issubclass(op, ast.boolop):
                    return ast.BoolOp(op=op, values=[left, right])
                elif issubclass(op, ast.cmpop):
                    return ast.Compare(left=left, ops=[op],
                                       comparators=[right])
                else:
                    return ast.BinOp(op=op, left=left, right=right)

        # f(x)
        elif self == LogoCode.prim_myfunction:
            param_asts = []
            for id_ in ['x', 'y', 'z'][:len(new_arg_asts) - 1]:
                param_asts.append(ast.Name(id=id_, ctx=ast.Param))
            func_ast = ast_extensions.LambdaWithStrBody(
                body_str=new_arg_asts[0].s, args=param_asts)
            return get_call_ast(func_ast, new_arg_asts[1:],
                                return_type=self.return_type)

        # square root
        elif self == Primitive.square_root:
            return get_call_ast('sqrt', new_arg_asts, new_kwarg_asts,
                                return_type=self.return_type)

        # random
        elif self in (Primitive.random_char, Primitive.random_int):
            uniform_ast = get_call_ast('uniform', new_arg_asts)
            round_ast = get_call_ast('round', [uniform_ast, ast.Num(n=0)])
            int_ast = get_call_ast('int', [round_ast], return_type=TYPE_INT)
            if self == Primitive.random_char:
                chr_ast = get_call_ast('chr', [int_ast], return_type=TYPE_CHAR)
                return chr_ast
            else:
                return int_ast

        # identity
        elif self == Primitive.identity:
            return new_arg_asts[0]

        # constant
        elif self == CONSTANTS.get:
            return TypedSubscript(value=ast.Name(id='CONSTANTS', ctx=ast.Load),
                                  slice_=ast.Index(value=new_arg_asts[0]),
                                  return_type=self.return_type)

        # group of Primitives or sandwich-clamp block
        elif self in (Primitive.group, LogoCode.prim_clamp):
            ast_list = []
            for prim in new_arg_asts[0]:
                if export_me(prim):
                    new_ast = value_to_ast(prim)
                    if isinstance(new_ast, ast.AST):
                        ast_list.append(new_ast)
            return ast_list

        # set turtle
        elif self == LogoCode.prim_turtle:
            text = 'turtle = turtles.get_active_turtle()'
            return [get_call_ast('logo.prim_turtle', new_arg_asts),
                    ast_extensions.ExtraCode(text)]

        elif self == LogoCode.active_turtle:
            text = 'turtle = turtles.get_active_turtle()'
            return ast_extensions.ExtraCode(text)

        # comment
        elif self == Primitive.comment:
            if isinstance(new_arg_asts[0], ast.Str):
                text = ' ' + str(new_arg_asts[0].s)
            else:
                text = ' ' + str(new_arg_asts[0])
            return ast_extensions.Comment(text)

        # print
        elif self == TurtleArtWindow.print_:
            func_name = self.get_name_for_export()
            call_ast = get_call_ast(func_name, new_arg_asts)
            print_ast = ast.Print(values=new_arg_asts[:1], dest=None, nl=True)
            return [call_ast, print_ast]

        # heap
        elif self == LogoCode.get_heap:
            return TypedName(id_='logo.heap', return_type=self.return_type)
        elif self == LogoCode.reset_heap:
            target_ast = ast.Name(id='logo.heap', ctx=ast.Store)
            value_ast = ast.List(elts=[], ctx=ast.Load)
            return ast.Assign(targets=[target_ast], value=value_ast)

        # NORMAL FUNCTION CALL #

        else:
            func_name = self.get_name_for_export()
            return get_call_ast(func_name, new_arg_asts, new_kwarg_asts,
                                return_type=self.return_type)

    def __eq__(self, other):
        """ Two Primitives are equal iff their all their properties are equal.
        Consider bound and unbound methods equal. """
        # other is a Primitive
        if isinstance(other, Primitive):
            return (self == other.func and
                    self.return_type == other.return_type and
                    self.arg_descs == other.arg_descs and
                    self.kwarg_descs == other.kwarg_descs and
                    self.call_afterwards == other.call_afterwards and
                    self.export_me == other.export_me)

        # other is a callable
        elif callable(other):
            if is_instancemethod(self.func) != is_instancemethod(other):
                return False
            elif is_instancemethod(self.func):  # and is_instancemethod(other):
                return (self.func.__self__.__class__ == other.__self__.__class__ and
                        self.func.__func__ == other.__func__)
            else:
                return self.func == other

        elif is_staticmethod(other):
            return self.func == other.__func__

        # other is neither a Primitive nor a callable
        else:
            return False

    def wants_turtle(self):
        """Does this Primitive want to get the active turtle as its first
        argument?"""
        return self._wants(Turtle)

    def wants_turtles(self):
        """ Does this Primitive want to get the Turtles instance as its
        first argument? """
        return self._wants(Turtles)

    def wants_canvas(self):
        """ Does this Primitive want to get the canvas as its first
        argument? """
        return self._wants(TurtleGraphics)

    def wants_logocode(self):
        """ Does this Primitive want to get the LogoCode instance as its
        first argument? """
        return (self.func.__name__ == '<lambda>' or self._wants(LogoCode))

    def wants_heap(self):
        """ Does this Primitive want to get the heap as its first argument? """
        return ((hasattr(self.func, '__self__') and
                 isinstance(self.func.__self__, list)) or
                self.func in list(list.__dict__.values()))

    def wants_tawindow(self):
        """ Does this Primitive want to get the TurtleArtWindow instance
        as its first argument? """
        return self._wants(TurtleArtWindow)

    def wants_plugin(self):
        """Does this Primitive want to get a plugin instance as its first
        argument? """
        for obj in list(global_objects.keys()):
            if self._wants(global_objects[obj].__class__):
                return True, obj
        return False, None

    def wants_nothing(self):
        """Does this Primitive want nothing as its first argument? I.e. does
        it want to be passed all the arguments of the block and
        nothing else?"""
        return not is_instancemethod(self.func)

    def _wants(self, theClass):
        try:
            return inspect.getattr_static(
                theClass, self.func.__name__).__class__.__name__ == 'function'
        except AttributeError:
            return False

    # treat the following methods in a special way when converting the
    # Primitive to an AST

    @staticmethod
    def controller_repeat(num):
        """ Loop controller for the 'repeat' block """
        for i in range(num):
            yield True
        yield False

    @staticmethod
    def controller_forever():
        """ Loop controller for the 'forever' block """
        while True:
            yield True

    @staticmethod
    def controller_while(condition):
        """ Loop controller for the 'while' block
        condition -- Primitive that is evaluated every time through the
            loop """
        condition.allow_call_args(recursive=True)
        while condition():
            yield True
        yield False

    @staticmethod
    def controller_until(condition):
        """ Loop controller for the 'until' block
        condition -- Primitive that is evaluated every time through the
            loop """
        condition.allow_call_args(recursive=True)
        while not condition():
            yield True
        yield False

    LOOP_CONTROLLERS = [controller_repeat, controller_forever,
                        controller_while, controller_until]

    def _get_loop_controller(self):
        """ Return the controller for this loop Primitive. Raise a
        ValueError if no controller was found. """
        def _is_loop_controller(candidate):
            return (callable(candidate) and
                    candidate in Primitive.LOOP_CONTROLLERS)

        for desc in self.arg_descs:
            if isinstance(desc, ConstantArg):
                value = desc.value
                if _is_loop_controller(value):
                    return value
            elif isinstance(desc, ArgSlot):
                wrapper = desc.wrapper
                if _is_loop_controller(wrapper):
                    return wrapper

        # no controller found
        raise PyExportError("found no loop controller for " + repr(self))

    @staticmethod
    def do_nothing():
        pass

    @staticmethod
    def identity(arg):
        """ Return the argument unchanged """
        return arg

    @staticmethod
    def group(prim_list):
        """ Group together multiple Primitives into one. Treat each Primitive
        as a separate line of code. """
        return_val = None
        for prim in prim_list:
            return_val = prim()
        return return_val

    @staticmethod
    def plus(arg1, arg2=None):
        """ If only one argument is given, prefix it with '+'. If two
        arguments are given, add the second to the first. If the first
        argument is a tuple of length 2 and the second is None, use the
        values in the tuple as arg1 and arg2. """
        if isinstance(arg1, (list, tuple)) and len(arg1) == 2 and arg2 is None:
            (arg1, arg2) = arg1
        if arg2 is None:
            return + arg1
        elif isinstance(arg1, Vector) and isinstance(arg2, Vector):
            vector = []
            for i in range(len(arg1.vector)):
                vector.append(arg1.vector[i] + arg2.vector[i])
            return Vector(arg1.name, vector)
        else:
            return arg1 + arg2

    @staticmethod
    def minus(arg1, arg2=None):
        """ If only one argument is given, change its sign. If two
        arguments are given, subtract the second from the first. """
        if arg2 is None:
            return - arg1
        elif isinstance(arg1, Vector) and isinstance(arg2, Vector):
            vector = []
            for i in range(len(arg1.vector)):
                vector.append(arg1.vector[i] - arg2.vector[i])
            return Vector(arg1.name, vector)
        else:
            return arg1 - arg2

    @staticmethod
    def multiply(arg1, arg2):
        """ Multiply the two arguments """
        if isinstance(arg1, Vector) and isinstance(arg2, (int, float)):
            vector = []
            for i in range(len(arg1.vector)):
                vector.append(arg1.vector[i] * arg2)
            return Vector(arg1.name, vector)
        elif isinstance(arg2, Vector) and isinstance(arg1, (int, float)):
            vector = []
            for i in range(len(arg2.vector)):
                vector.append(arg2.vector[i] * arg1)
            return Vector(arg2.name, vector)
        else:
            return arg1 * arg2

    @staticmethod
    def divide(arg1, arg2):
        """ Divide the first argument by the second """
        if arg2 == 0:
            raise logoerror("#zerodivide")

        if isinstance(arg1, Vector) and isinstance(arg2, (int, float)):
            vector = []
            for i in range(len(arg1.vector)):
                vector.append(arg1.vector[i] / arg2)
            return Vector(arg1.name, vector)
        elif isinstance(arg2, Vector) and isinstance(arg1, (int, float)):
            vector = []
            for i in range(len(arg2.vector)):
                vector.append(arg2.vector[i] / arg1)
            return Vector(arg2.name, vector)
        else:
            return float(arg1) / arg2

    @staticmethod
    def modulo(arg1, arg2):
        """ Return the remainder of dividing the first argument by the second.
        If the first argument is a string, format it with the value(s) in
        the second argument. """
        return arg1 % arg2

    @staticmethod
    def power(arg1, arg2):
        """ Raise the first argument to the power given by the second """
        return arg1 ** arg2

    @staticmethod
    def square_root(arg1):
        """ Return the square root of the argument. If it is a negative
        number, raise a NegativeRootError. """
        if arg1 < 0:
            raise NegativeRootError(neg_value=arg1)
        return sqrt(arg1)

    @staticmethod
    def and_(arg1, arg2):
        """ Logcially conjoin the two arguments (using short-circuting) """
        return arg1 and arg2

    @staticmethod
    def or_(arg1, arg2):
        """ Logically disjoin the two arguments (using short-circuting) """
        return arg1 or arg2

    @staticmethod
    def not_(arg):
        """ Return True if the argument evaluates to False, and False
        otherwise. """
        return not arg

    @staticmethod
    def equals(arg1, arg2):
        """ Return arg1 == arg2 """
        # See comment in tatype.py TYPE_BOX -> TYPE_COLOR
        if isinstance(arg1, ColorObj) or isinstance(arg2, ColorObj):
            return str(arg1) == str(arg2)
        else:
            return arg1 == arg2

    @staticmethod
    def less(arg1, arg2):
        """ Return arg1 < arg2 """
        # See comment in tatype.py TYPE_BOX -> TYPE_COLOR
        if isinstance(arg1, ColorObj) or isinstance(arg2, ColorObj):
            return float(arg1) < float(arg2)
        else:
            return arg1 < arg2

    @staticmethod
    def greater(arg1, arg2):
        """ Return arg1 > arg2 """
        # See comment in tatype.py TYPE_BOX -> TYPE_COLOR
        if isinstance(arg1, ColorObj) or isinstance(arg2, ColorObj):
            return float(arg1) > float(arg2)
        else:
            return arg1 > arg2

    @staticmethod
    def comment(text):
        """In 'snail' execution mode, display the comment. Else, do
        nothing."""
        tw = global_objects["window"]
        if not tw.hide and tw.step_time != 0:
            tw.showlabel('print', text)

    @staticmethod
    def random_int(lower, upper):
        """ Choose a random integer between lower and upper, which must be
        integers """
        return int(round(uniform(lower, upper), 0))

    @staticmethod
    def random_char(lower, upper):
        """ Choose a random Unicode code point between lower and upper,
        which must be integers """
        return chr(Primitive.random_int(lower, upper))


class Disjunction(tuple):

    """ Abstract disjunction class (not to be instantiated directly) """

    def __init__(self, iterable):
        self = tuple(iterable)

    def __repr__(self):
        s = ["("]
        for disj in self:
            s.append(repr(disj))
            s.append(" or ")
        s.pop()
        s.append(")")
        return "".join(s)

    def get_alternatives(self):
        """ Return a tuple of alternatives, i.e. self """
        return self


class PrimitiveDisjunction(Disjunction, Primitive):

    """ Disjunction of two or more Primitives. PrimitiveDisjunctions may not
    be nested. """

    @property
    def return_type(self):
        """ Tuple of the return_types of all disjuncts """
        return TypeDisjunction((prim.return_type for prim in self))

    def __call__(self, *runtime_args, **runtime_kwargs):
        """ Loop over the disjunct Primitives and try to fill their slots
        with the given args and kwargs. Call the first Primitives whose
        slots could be filled successfully. If all disjunct Primitives
        fail, raise the last error that occurred. """

        # remove the first argument if it is a LogoCode instance
        if runtime_args and isinstance(runtime_args[0], LogoCode):
            runtime_args = runtime_args[1:]

        error = None
        for prim in self:
            try:
                new_prim = prim.fill_slots(runtime_args, runtime_kwargs,
                                           convert_to_ast=False)
            except TATypeError as error:
                # on failure, try the next one
                continue
            else:
                # on success, call this Primitive
                return new_prim()

        # if we get here, all disjuncts failed
        if error is not None:
            raise error


class ArgListDisjunction(Disjunction):

    """ Disjunction of two or more argument lists """
    pass


class ArgSlot(object):

    """ Description of the requirements that a Primitive demands from an
    argument or keyword argument. An ArgSlot is filled at runtime, based
    on the block program structure. """

    def __init__(self, type_, call_arg=True, wrapper=None):
        """
        type_ -- what type of the type hierarchy the argument should have
            (after the wrapper has been applied)
        call_arg -- if this argument is callable, should it be called and
            its return value passed to the parent Primitive (True, the
            default), or should it be passed as it is (False)?
        wrapper -- a Primitive that is 'wrapped around' the argument before
            it gets passed to its parent Primitive. Wrappers can be nested
            infinitely. """
        self.type = type_
        self.call_arg = call_arg
        self.wrapper = wrapper

    def __repr__(self):
        s = ["ArgSlot(type="]
        s.append(repr(self.type))
        if not self.call_arg:
            s.append(", call=")
            s.append(repr(self.call_arg))
        if self.wrapper is not None:
            s.append(", wrapper=")
            s.append(repr(self.wrapper))
        s.append(")")
        return "".join(s)

    def get_alternatives(self):
        """ Return a tuple of slot alternatives, i.e. (self, ) """
        return (self, )

    def fill(self, argument, convert_to_ast=False, call_my_args=True):
        """ Try to fill this argument slot with the given argument. Return
        a ConstantArg containing the result. If there is a type problem,
        raise a TATypeError. """
        if isinstance(argument, ast.AST):
            convert_to_ast = True

        # 1. can the argument be called?
        (func_disjunction, args) = (None, [])
        if (isinstance(argument, tuple) and argument and
                callable(argument[0])):
            func_disjunction = argument[0]
            if len(argument) >= 2 and isinstance(argument[1], LogoCode):
                args = argument[2:]
            else:
                args = argument[1:]
        elif callable(argument):
            func_disjunction = argument

        # make sure we can loop over func_disjunction
        if not isinstance(func_disjunction, PrimitiveDisjunction):
            func_disjunction = PrimitiveDisjunction((func_disjunction, ))

        error = None
        bad_value = argument  # the value that caused the TATypeError
        for func in func_disjunction:
            error = None
            for slot in self.get_alternatives():

                if isinstance(slot.wrapper, PrimitiveDisjunction):
                    wrapper_disjunction = slot.wrapper
                else:
                    wrapper_disjunction = PrimitiveDisjunction((slot.wrapper,))

                for wrapper in wrapper_disjunction:

                    # check if the argument can fill this slot (type-wise)
                    # (lambda functions are always accepted)
                    if getattr(func, '__name__', None) == '<lambda>':
                        converter = identity
                        old_type = TYPE_OBJECT
                        new_type = slot.type
                    else:
                        if wrapper is not None:
                            arg_types = get_type(wrapper)[0]
                            bad_value = wrapper
                        elif func is not None:
                            arg_types = get_type(func)[0]
                            bad_value = func
                        else:
                            arg_types = get_type(argument)[0]
                            bad_value = argument
                        converter = None
                        if not isinstance(arg_types, TypeDisjunction):
                            arg_types = TypeDisjunction((arg_types, ))
                        if isinstance(slot.type, TypeDisjunction):
                            slot_types = slot.type
                        else:
                            slot_types = TypeDisjunction((slot.type, ))
                        for old_type in arg_types:
                            for new_type in slot_types:
                                converter = get_converter(old_type, new_type)
                                if converter is not None:
                                    break
                            if converter is not None:
                                break
                        # unable to convert, try next wrapper/ slot/ func
                        if converter is None:
                            continue

                    # 1. (cont'd) call the argument or pass it on as a callable
                    called_argument = argument
                    if func is not None:
                        func_prim = func
                        if not isinstance(func_prim, Primitive):
                            func_prim = Primitive(
                                func_prim,
                                [ArgSlot(TYPE_OBJECT)] * len(args))
                        try:
                            func_prim = func_prim.fill_slots(
                                args,
                                convert_to_ast=convert_to_ast,
                                call_my_args=(slot.call_arg and call_my_args))
                        except TATypeError as e:
                            error = e
                            if Primitive._DEBUG:
                                traceback.print_exc()
                            # on failure, try next wrapper/ slot/ func
                            bad_value = error.bad_value
                            continue
                        if convert_to_ast:
                            called_argument = func_prim.get_ast()
                        else:
                            if slot.call_arg and call_my_args:
                                # call and pass on the return value
                                called_argument = func_prim()
                            else:
                                # don't call and pass on the callable
                                called_argument = func_prim

                    # 2. apply any wrappers
                    wrapped_argument = called_argument
                    if wrapper is not None:
                        if convert_to_ast:
                            if not hasattr(wrapper, "get_ast"):
                                raise PyExportError(
                                    ("cannot convert callable"
                                     " %s to an AST") % (repr(wrapper)))
                            wrapped_argument = wrapper.get_ast(
                                called_argument)
                        else:
                            if slot.call_arg and call_my_args:
                                wrapped_argument = wrapper(called_argument)
                            else:
                                wrapped_argument = wrapper.fill_slots(
                                    [called_argument], call_my_args=False)

                    # last chance to convert raw values to ASTs
                    # (but not lists of ASTs)
                    if (convert_to_ast and
                            not isinstance(wrapped_argument, ast.AST) and
                            not (isinstance(wrapped_argument, list) and
                                 wrapped_argument and
                                 isinstance(wrapped_argument[0], ast.AST))):
                        wrapped_argument = value_to_ast(wrapped_argument)

                    # 3. check the type and convert the argument if necessary
                    converted_argument = wrapped_argument
                    if slot.call_arg and call_my_args:
                        try:
                            converted_argument = convert(
                                wrapped_argument,
                                new_type, old_type=old_type,
                                converter=converter)
                        except TATypeError as e:
                            error = e
                            if Primitive._DEBUG:
                                traceback.print_exc()
                            # on failure, try next wrapper/ slot/ func
                            bad_value = wrapped_argument
                            continue
                    elif converter != identity:
                        converted_argument = Primitive(
                            converter,
                            return_type=new_type,
                            arg_descs=[ConstantArg(wrapped_argument,
                                                   value_type=old_type,
                                                   call_arg=False)])
                    # on success, return the result
                    return ConstantArg(
                        converted_argument,
                        value_type=new_type,
                        call_arg=(slot.call_arg and call_my_args))

        # if we haven't returned anything yet, then all alternatives failed
        if error is not None:
            raise error
        else:
            raise TATypeError(bad_value=bad_value, bad_type=old_type,
                              req_type=new_type)


class ArgSlotDisjunction(Disjunction, ArgSlot):

    """ Disjunction of two or more argument slots """
    pass


class ConstantArg(object):

    """ A constant argument or keyword argument to a Primitive. It is
    independent of the block program structure. """

    def __init__(self, value, call_arg=True, value_type=None):
        """ call_arg -- call the value before returning it?
        value_type -- the type of the value (from the TA type system). This
            is useful to store e.g., the return type of call ASTs. """
        self.value = value
        self.call_arg = call_arg
        self.value_type = value_type

    def get(self, convert_to_ast=False):
        """ If call_arg is True and the value is callable, call the value
        and return its return value. Else, return the value unchanged.
        convert_to_ast -- return the equivalent AST instead of a raw value """
        if self.call_arg and callable(self.value):
            if convert_to_ast:
                return value_to_ast(self.value)
            else:
                return self.value()
        else:
            if convert_to_ast and not isinstance(self.value, list):
                return value_to_ast(self.value)
            else:
                return self.value

    def get_value_type(self):
        """ If this ConstantArg has stored the type of its value, return
        that. Else, use get_type(...) to guess the type of the value. """
        if self.value_type is None:
            return get_type(self.value)[0]
        else:
            return self.value_type

    def __repr__(self):
        s = ["ConstantArg("]
        s.append(repr(self.value))
        if not self.call_arg:
            s.append(", call=")
            s.append(repr(self.call_arg))
        s.append(")")
        return "".join(s)


def or_(*disjuncts):
    """ Return a disjunction object of the same type as the disjuncts. If
    the item type cannot be linked to a Disjunction class, return a tuple
    of the disjuncts. """
    if isinstance(disjuncts[0], Primitive):
        return PrimitiveDisjunction(disjuncts)
    elif isinstance(disjuncts[0], (list, ArgListDisjunction)):
        return ArgListDisjunction(disjuncts)
    elif isinstance(disjuncts[0], ArgSlot):
        return ArgSlotDisjunction(disjuncts)
    elif isinstance(disjuncts[0], Type):
        return TypeDisjunction(disjuncts)
    else:
        return tuple(disjuncts)


def value_to_ast(value, *args_for_prim, **kwargs_for_prim):
    """ Turn a value into an AST. Supported types: Primitive, int, float,
    bool, basestring, list
    If the value is already an AST, return it unchanged.
    If the value is a non-exportable Primitive, return None. """
    # already an AST
    if isinstance(value, ast.AST):
        return value
    # Primitive
    elif isinstance(value, Primitive):
        if value.export_me:
            return value.get_ast(*args_for_prim, **kwargs_for_prim)
        else:
            return None
    # boolean
    elif isinstance(value, bool):
        return ast.Name(id=str(value), ctx=ast.Load)
    # number
    elif isinstance(value, (int, float)):
        return ast.Num(n=value)
    # string
    elif isinstance(value, str):
        return ast.Str(value)
    # list (recursively transform to an AST)
    elif isinstance(value, list):
        ast_list = []
        for item in value:
            item_ast = value_to_ast(item)
            if item_ast is not None:
                ast_list.append(item_ast)
        return ast.List(elts=ast_list, ctx=ast.Load)
    # color
    elif isinstance(value, Color):
        # call to the Color constructor with this object's values,
        # e.g., Color('red', 0, 50, 100)
        return get_call_ast('Color', [value.name, value.color,
                                      value.shade, value.gray],
                            return_type=TYPE_COLOR)
    # vector
    elif isinstance(value, Vector):
        # call to the Vector constructor with this object's values,
        # e.g., Vector('banana', [105, 1, 27, 3, 0])
        return get_call_ast('Vector', [value.name, value.vector],
                            return_type=TYPE_VECTOR)
    # media
    elif isinstance(value, Media):
        args = [value_to_ast(value.type), value_to_ast(value.value)]
        return get_call_ast('Media', args, return_type=TYPE_MEDIA)
    # unknown
    else:
        raise PyExportError("unknown type of raw value: " + repr(type(value)))


def ast_yield_true():
    return ast.Yield(value=ast.Name(id='True', ctx=ast.Load))


def export_me(something):
    """ Return True iff this is not a Primitive or its export_me attribute
    is True, i.e. everything is exportable except for Primitives with
    export_me == False """
    return not isinstance(something, Primitive) or something.export_me



# -*- coding: utf-8 -*-
# Copyright (C) 2006-2012 Søren Roug, European Environment Agency
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Contributor(s):
#
from namespaces import *
import re
import types

pattern_color = re.compile(r'#[0-9a-fA-F]{6}')
pattern_vector3D = re.compile(
    r'\([ ]*-?([0-9]+(\.[0-9]*)?|\.[0-9]+)([ ]+-?([0-9]+(\.[0-9]*)?|\.[0-9]+)){2}[ ]*\)')


def make_NCName(arg):
    for c in (':', ' '):
        arg = arg.replace(c, "_%x_" % ord(c))
    return arg


def cnv_anyURI(attribute, arg, element):
    return unicode(arg)


def cnv_boolean(attribute, arg, element):
    """ XML Schema Part 2: Datatypes Second Edition
        An instance of a datatype that is defined as boolean can have the
        following legal literals {true, false, 1, 0}
    """
    if str(arg).lower() in ("0", "false", "no"):
        return "false"
    if str(arg).lower() in ("1", "true", "yes"):
        return "true"
    raise ValueError("'%s' not allowed as Boolean value for %s" % (
        str(arg), attribute))

# Potentially accept color values


def cnv_color(attribute, arg, element):
    """ A RGB color in conformance with §5.9.11 of [XSL], that is a RGB color in notation “#rrggbb”, where
        rr, gg and bb are 8-bit hexadecimal digits.
    """
    return str(arg)


def cnv_configtype(attribute, arg, element):
    if str(arg) not in ("boolean", "short", "int", "long",
                        "double", "string", "datetime", "base64Binary"):
        raise ValueError("'%s' not allowed" % str(arg))
    return str(arg)


def cnv_data_source_has_labels(attribute, arg, element):
    if str(arg) not in ("none", "row", "column", "both"):
        raise ValueError("'%s' not allowed" % str(arg))
    return str(arg)

# Understand different date formats


def cnv_date(attribute, arg, element):
    """ A dateOrDateTime value is either an [xmlschema-2] date value or an [xmlschema-2] dateTime
        value.
    """
    return str(arg)


def cnv_dateTime(attribute, arg, element):
    """ A dateOrDateTime value is either an [xmlschema-2] date value or an [xmlschema-2] dateTime
        value.
    """
    return str(arg)


def cnv_double(attribute, arg, element):
    return str(arg)


def cnv_duration(attribute, arg, element):
    return str(arg)


def cnv_family(attribute, arg, element):
    """ A style family """
    if str(arg) not in (
        "text",
        "paragraph",
        "section",
        "ruby",
        "table",
        "table-column",
        "table-row",
        "table-cell",
        "graphic",
        "presentation",
        "drawing-page",
            "chart"):
        raise ValueError("'%s' not allowed" % str(arg))
    return str(arg)


def __save_prefix(attribute, arg, element):
    prefix = arg.split(':', 1)[0]
    if prefix == arg:
        return unicode(arg)
    namespace = element.get_knownns(prefix)
    if namespace is None:
        #raise ValueError, "'%s' is an unknown prefix" % str(prefix)
        return unicode(arg)
    p = element.get_nsprefix(namespace)
    return unicode(arg)


def cnv_formula(attribute, arg, element):
    """ A string containing a formula. Formulas do not have a predefined syntax, but the string should
        begin with a namespace prefix, followed by a “:” (COLON, U+003A) separator, followed by the text
        of the formula. The namespace bound to the prefix determines the syntax and semantics of the
        formula.
    """
    return __save_prefix(attribute, arg, element)


def cnv_ID(attribute, arg, element):
    return str(arg)


def cnv_IDREF(attribute, arg, element):
    return str(arg)


def cnv_integer(attribute, arg, element):
    return str(arg)


def cnv_legend_position(attribute, arg, element):
    if str(arg) not in ("start", "end", "top", "bottom",
                        "top-start", "bottom-start", "top-end", "bottom-end"):
        raise ValueError("'%s' not allowed" % str(arg))
    return str(arg)


pattern_length = re.compile(
    r'-?([0-9]+(\.[0-9]*)?|\.[0-9]+)((cm)|(mm)|(in)|(pt)|(pc)|(px))')


def cnv_length(attribute, arg, element):
    """ A (positive or negative) physical length, consisting of magnitude and unit, in conformance with the
        Units of Measure defined in §5.9.13 of [XSL].
    """
    global pattern_length
    if not pattern_length.match(arg):
        raise ValueError("'%s' is not a valid length" % arg)
    return arg


def cnv_lengthorpercent(attribute, arg, element):
    failed = False
    try:
        return cnv_length(attribute, arg, element)
    except BaseException:
        failed = True
    try:
        return cnv_percent(attribute, arg, element)
    except BaseException:
        failed = True
    if failed:
        raise ValueError("'%s' is not a valid length or percent" % arg)
    return arg


def cnv_metavaluetype(attribute, arg, element):
    if str(arg) not in ("float", "date", "time", "boolean", "string"):
        raise ValueError("'%s' not allowed" % str(arg))
    return str(arg)


def cnv_major_minor(attribute, arg, element):
    if arg not in ('major', 'minor'):
        raise ValueError("'%s' is not either 'minor' or 'major'" % arg)


pattern_namespacedToken = re.compile(r'[0-9a-zA-Z_]+:[0-9a-zA-Z._\-]+')


def cnv_namespacedToken(attribute, arg, element):
    global pattern_namespacedToken

    if not pattern_namespacedToken.match(arg):
        raise ValueError("'%s' is not a valid namespaced token" % arg)
    return __save_prefix(attribute, arg, element)


def cnv_NCName(attribute, arg, element):
    """ NCName is defined in http://www.w3.org/TR/REC-xml-names/#NT-NCName
        Essentially an XML name minus ':'
    """
    if type(arg) in types.StringTypes:
        return make_NCName(arg)
    else:
        return arg.getAttrNS(STYLENS, 'name')

# This function takes either an instance of a style (preferred)
# or a text string naming the style. If it is a text string, then it must
# already have been converted to an NCName
# The text-string argument is mainly for when we build a structure from XML


def cnv_StyleNameRef(attribute, arg, element):
    try:
        return arg.getAttrNS(STYLENS, 'name')
    except BaseException:
        return arg

# This function takes either an instance of a style (preferred)
# or a text string naming the style. If it is a text string, then it must
# already have been converted to an NCName
# The text-string argument is mainly for when we build a structure from XML


def cnv_DrawNameRef(attribute, arg, element):
    try:
        return arg.getAttrNS(DRAWNS, 'name')
    except BaseException:
        return arg

# Must accept list of Style objects


def cnv_NCNames(attribute, arg, element):
    return ' '.join(arg)


def cnv_nonNegativeInteger(attribute, arg, element):
    return str(arg)


pattern_percent = re.compile(r'-?([0-9]+(\.[0-9]*)?|\.[0-9]+)%')


def cnv_percent(attribute, arg, element):
    global pattern_percent
    if not pattern_percent.match(arg):
        raise ValueError("'%s' is not a valid length" % arg)
    return arg


# Real one doesn't allow floating point values
pattern_points = re.compile(r'-?[0-9]+,-?[0-9]+([ ]+-?[0-9]+,-?[0-9]+)*')
#pattern_points = re.compile(r'-?[0-9.]+,-?[0-9.]+([ ]+-?[0-9.]+,-?[0-9.]+)*')


def cnv_points(attribute, arg, element):
    global pattern_points
    if type(arg) in types.StringTypes:
        if not pattern_points.match(arg):
            raise ValueError(
                "x,y are separated by a comma and the points are separated by white spaces")
        return arg
    else:
        try:
            strarg = ' '.join(["%d,%d" % p for p in arg])
        except BaseException:
            raise ValueError(
                "Points must be string or [(0,0),(1,1)] - not %s" %
                arg)
        return strarg


def cnv_positiveInteger(attribute, arg, element):
    return str(arg)


def cnv_string(attribute, arg, element):
    return unicode(arg)


def cnv_textnoteclass(attribute, arg, element):
    if str(arg) not in ("footnote", "endnote"):
        raise ValueError("'%s' not allowed" % str(arg))
    return str(arg)

# Understand different time formats


def cnv_time(attribute, arg, element):
    return str(arg)


def cnv_token(attribute, arg, element):
    return str(arg)


pattern_viewbox = re.compile(r'-?[0-9]+([ ]+-?[0-9]+){3}$')


def cnv_viewbox(attribute, arg, element):
    global pattern_viewbox
    if not pattern_viewbox.match(arg):
        raise ValueError(
            "viewBox must be four integers separated by whitespaces")
    return arg


def cnv_xlinkshow(attribute, arg, element):
    if str(arg) not in ("new", "replace", "embed"):
        raise ValueError("'%s' not allowed" % str(arg))
    return str(arg)


attrconverters = {
    ((ANIMNS, u'audio-level'), None): cnv_double,
    ((ANIMNS, u'color-interpolation'), None): cnv_string,
    ((ANIMNS, u'color-interpolation-direction'), None): cnv_string,
    ((ANIMNS, u'command'), None): cnv_string,
    ((ANIMNS, u'formula'), None): cnv_string,
    ((ANIMNS, u'id'), None): cnv_ID,
    ((ANIMNS, u'iterate-interval'), None): cnv_duration,
    ((ANIMNS, u'iterate-type'), None): cnv_string,
    ((ANIMNS, u'name'), None): cnv_string,
    ((ANIMNS, u'sub-item'), None): cnv_string,
    ((ANIMNS, u'value'), None): cnv_string,
    #	((DBNS,u'type'), None): cnv_namespacedToken,
    ((CHARTNS, u'attached-axis'), None): cnv_string,
    ((CHARTNS, u'class'), (CHARTNS, u'grid')): cnv_major_minor,
    ((CHARTNS, u'class'), None): cnv_namespacedToken,
    ((CHARTNS, u'column-mapping'), None): cnv_string,
    ((CHARTNS, u'connect-bars'), None): cnv_boolean,
    ((CHARTNS, u'data-label-number'), None): cnv_string,
    ((CHARTNS, u'data-label-symbol'), None): cnv_boolean,
    ((CHARTNS, u'data-label-text'), None): cnv_boolean,
    ((CHARTNS, u'data-source-has-labels'), None): cnv_data_source_has_labels,
    ((CHARTNS, u'deep'), None): cnv_boolean,
    ((CHARTNS, u'dimension'), None): cnv_string,
    ((CHARTNS, u'display-label'), None): cnv_boolean,
    ((CHARTNS, u'error-category'), None): cnv_string,
    ((CHARTNS, u'error-lower-indicator'), None): cnv_boolean,
    ((CHARTNS, u'error-lower-limit'), None): cnv_string,
    ((CHARTNS, u'error-margin'), None): cnv_string,
    ((CHARTNS, u'error-percentage'), None): cnv_string,
    ((CHARTNS, u'error-upper-indicator'), None): cnv_boolean,
    ((CHARTNS, u'error-upper-limit'), None): cnv_string,
    ((CHARTNS, u'gap-width'), None): cnv_string,
    ((CHARTNS, u'interpolation'), None): cnv_string,
    ((CHARTNS, u'interval-major'), None): cnv_string,
    ((CHARTNS, u'interval-minor-divisor'), None): cnv_string,
    ((CHARTNS, u'japanese-candle-stick'), None): cnv_boolean,
    ((CHARTNS, u'label-arrangement'), None): cnv_string,
    ((CHARTNS, u'label-cell-address'), None): cnv_string,
    ((CHARTNS, u'legend-align'), None): cnv_string,
    ((CHARTNS, u'legend-position'), None): cnv_legend_position,
    ((CHARTNS, u'lines'), None): cnv_boolean,
    ((CHARTNS, u'link-data-style-to-source'), None): cnv_boolean,
    ((CHARTNS, u'logarithmic'), None): cnv_boolean,
    ((CHARTNS, u'maximum'), None): cnv_string,
    ((CHARTNS, u'mean-value'), None): cnv_boolean,
    ((CHARTNS, u'minimum'), None): cnv_string,
    ((CHARTNS, u'name'), None): cnv_string,
    ((CHARTNS, u'origin'), None): cnv_string,
    ((CHARTNS, u'overlap'), None): cnv_string,
    ((CHARTNS, u'percentage'), None): cnv_boolean,
    ((CHARTNS, u'pie-offset'), None): cnv_string,
    ((CHARTNS, u'regression-type'), None): cnv_string,
    ((CHARTNS, u'repeated'), None): cnv_nonNegativeInteger,
    ((CHARTNS, u'row-mapping'), None): cnv_string,
    ((CHARTNS, u'scale-text'), None): cnv_boolean,
    ((CHARTNS, u'series-source'), None): cnv_string,
    ((CHARTNS, u'solid-type'), None): cnv_string,
    ((CHARTNS, u'spline-order'), None): cnv_string,
    ((CHARTNS, u'spline-resolution'), None): cnv_string,
    ((CHARTNS, u'stacked'), None): cnv_boolean,
    ((CHARTNS, u'style-name'), None): cnv_StyleNameRef,
    ((CHARTNS, u'symbol-height'), None): cnv_string,
    ((CHARTNS, u'symbol-name'), None): cnv_string,
    ((CHARTNS, u'symbol-type'), None): cnv_string,
    ((CHARTNS, u'symbol-width'), None): cnv_string,
    ((CHARTNS, u'text-overlap'), None): cnv_boolean,
    ((CHARTNS, u'three-dimensional'), None): cnv_boolean,
    ((CHARTNS, u'tick-marks-major-inner'), None): cnv_boolean,
    ((CHARTNS, u'tick-marks-major-outer'), None): cnv_boolean,
    ((CHARTNS, u'tick-marks-minor-inner'), None): cnv_boolean,
    ((CHARTNS, u'tick-marks-minor-outer'), None): cnv_boolean,
    ((CHARTNS, u'values-cell-range-address'), None): cnv_string,
    ((CHARTNS, u'vertical'), None): cnv_boolean,
    ((CHARTNS, u'visible'), None): cnv_boolean,
    ((CONFIGNS, u'name'), None): cnv_formula,
    ((CONFIGNS, u'type'), None): cnv_configtype,
    ((DR3DNS, u'ambient-color'), None): cnv_string,
    ((DR3DNS, u'back-scale'), None): cnv_string,
    ((DR3DNS, u'backface-culling'), None): cnv_string,
    ((DR3DNS, u'center'), None): cnv_string,
    ((DR3DNS, u'close-back'), None): cnv_boolean,
    ((DR3DNS, u'close-front'), None): cnv_boolean,
    ((DR3DNS, u'depth'), None): cnv_length,
    ((DR3DNS, u'diffuse-color'), None): cnv_string,
    ((DR3DNS, u'direction'), None): cnv_string,
    ((DR3DNS, u'distance'), None): cnv_length,
    ((DR3DNS, u'edge-rounding'), None): cnv_string,
    ((DR3DNS, u'edge-rounding-mode'), None): cnv_string,
    ((DR3DNS, u'emissive-color'), None): cnv_string,
    ((DR3DNS, u'enabled'), None): cnv_boolean,
    ((DR3DNS, u'end-angle'), None): cnv_string,
    ((DR3DNS, u'focal-length'), None): cnv_length,
    ((DR3DNS, u'horizontal-segments'), None): cnv_string,
    ((DR3DNS, u'lighting-mode'), None): cnv_boolean,
    ((DR3DNS, u'max-edge'), None): cnv_string,
    ((DR3DNS, u'min-edge'), None): cnv_string,
    ((DR3DNS, u'normals-direction'), None): cnv_string,
    ((DR3DNS, u'normals-kind'), None): cnv_string,
    ((DR3DNS, u'projection'), None): cnv_string,
    ((DR3DNS, u'shade-mode'), None): cnv_string,
    ((DR3DNS, u'shadow'), None): cnv_string,
    ((DR3DNS, u'shadow-slant'), None): cnv_nonNegativeInteger,
    ((DR3DNS, u'shininess'), None): cnv_string,
    ((DR3DNS, u'size'), None): cnv_string,
    ((DR3DNS, u'specular'), None): cnv_boolean,
    ((DR3DNS, u'specular-color'), None): cnv_string,
    ((DR3DNS, u'texture-filter'), None): cnv_string,
    ((DR3DNS, u'texture-generation-mode-x'), None): cnv_string,
    ((DR3DNS, u'texture-generation-mode-y'), None): cnv_string,
    ((DR3DNS, u'texture-kind'), None): cnv_string,
    ((DR3DNS, u'texture-mode'), None): cnv_string,
    ((DR3DNS, u'transform'), None): cnv_string,
    ((DR3DNS, u'vertical-segments'), None): cnv_string,
    ((DR3DNS, u'vpn'), None): cnv_string,
    ((DR3DNS, u'vrp'), None): cnv_string,
    ((DR3DNS, u'vup'), None): cnv_string,
    ((DRAWNS, u'align'), None): cnv_string,
    ((DRAWNS, u'angle'), None): cnv_integer,
    ((DRAWNS, u'archive'), None): cnv_string,
    ((DRAWNS, u'auto-grow-height'), None): cnv_boolean,
    ((DRAWNS, u'auto-grow-width'), None): cnv_boolean,
    ((DRAWNS, u'background-size'), None): cnv_string,
    ((DRAWNS, u'blue'), None): cnv_string,
    ((DRAWNS, u'border'), None): cnv_string,
    ((DRAWNS, u'caption-angle'), None): cnv_string,
    ((DRAWNS, u'caption-angle-type'), None): cnv_string,
    ((DRAWNS, u'caption-escape'), None): cnv_string,
    ((DRAWNS, u'caption-escape-direction'), None): cnv_string,
    ((DRAWNS, u'caption-fit-line-length'), None): cnv_boolean,
    ((DRAWNS, u'caption-gap'), None): cnv_string,
    ((DRAWNS, u'caption-line-length'), None): cnv_length,
    ((DRAWNS, u'caption-point-x'), None): cnv_string,
    ((DRAWNS, u'caption-point-y'), None): cnv_string,
    ((DRAWNS, u'caption-id'), None): cnv_IDREF,
    ((DRAWNS, u'caption-type'), None): cnv_string,
    ((DRAWNS, u'chain-next-name'), None): cnv_string,
    ((DRAWNS, u'class-id'), None): cnv_string,
    ((DRAWNS, u'class-names'), None): cnv_NCNames,
    ((DRAWNS, u'code'), None): cnv_string,
    ((DRAWNS, u'color'), None): cnv_string,
    ((DRAWNS, u'color-inversion'), None): cnv_boolean,
    ((DRAWNS, u'color-mode'), None): cnv_string,
    ((DRAWNS, u'concave'), None): cnv_string,
    ((DRAWNS, u'concentric-gradient-fill-allowed'), None): cnv_boolean,
    ((DRAWNS, u'contrast'), None): cnv_string,
    ((DRAWNS, u'control'), None): cnv_IDREF,
    ((DRAWNS, u'copy-of'), None): cnv_string,
    ((DRAWNS, u'corner-radius'), None): cnv_length,
    ((DRAWNS, u'corners'), None): cnv_positiveInteger,
    ((DRAWNS, u'cx'), None): cnv_string,
    ((DRAWNS, u'cy'), None): cnv_string,
    ((DRAWNS, u'data'), None): cnv_string,
    ((DRAWNS, u'decimal-places'), None): cnv_string,
    ((DRAWNS, u'display'), None): cnv_string,
    ((DRAWNS, u'display-name'), None): cnv_string,
    ((DRAWNS, u'distance'), None): cnv_lengthorpercent,
    ((DRAWNS, u'dots1'), None): cnv_integer,
    ((DRAWNS, u'dots1-length'), None): cnv_lengthorpercent,
    ((DRAWNS, u'dots2'), None): cnv_integer,
    ((DRAWNS, u'dots2-length'), None): cnv_lengthorpercent,
    ((DRAWNS, u'end-angle'), None): cnv_double,
    ((DRAWNS, u'end'), None): cnv_string,
    ((DRAWNS, u'end-color'), None): cnv_string,
    ((DRAWNS, u'end-glue-point'), None): cnv_nonNegativeInteger,
    ((DRAWNS, u'end-guide'), None): cnv_length,
    ((DRAWNS, u'end-intensity'), None): cnv_string,
    ((DRAWNS, u'end-line-spacing-horizontal'), None): cnv_string,
    ((DRAWNS, u'end-line-spacing-vertical'), None): cnv_string,
    ((DRAWNS, u'end-shape'), None): cnv_IDREF,
    ((DRAWNS, u'engine'), None): cnv_namespacedToken,
    ((DRAWNS, u'enhanced-path'), None): cnv_string,
    ((DRAWNS, u'escape-direction'), None): cnv_string,
    ((DRAWNS, u'extrusion-allowed'), None): cnv_boolean,
    ((DRAWNS, u'extrusion-brightness'), None): cnv_string,
    ((DRAWNS, u'extrusion'), None): cnv_boolean,
    ((DRAWNS, u'extrusion-color'), None): cnv_boolean,
    ((DRAWNS, u'extrusion-depth'), None): cnv_double,
    ((DRAWNS, u'extrusion-diffusion'), None): cnv_string,
    ((DRAWNS, u'extrusion-first-light-direction'), None): cnv_string,
    ((DRAWNS, u'extrusion-first-light-harsh'), None): cnv_boolean,
    ((DRAWNS, u'extrusion-first-light-level'), None): cnv_string,
    ((DRAWNS, u'extrusion-light-face'), None): cnv_boolean,
    ((DRAWNS, u'extrusion-metal'), None): cnv_boolean,
    ((DRAWNS, u'extrusion-number-of-line-segments'), None): cnv_integer,
    ((DRAWNS, u'extrusion-origin'), None): cnv_double,
    ((DRAWNS, u'extrusion-rotation-angle'), None): cnv_double,
    ((DRAWNS, u'extrusion-rotation-center'), None): cnv_string,
    ((DRAWNS, u'extrusion-second-light-direction'), None): cnv_string,
    ((DRAWNS, u'extrusion-second-light-harsh'), None): cnv_boolean,
    ((DRAWNS, u'extrusion-second-light-level'), None): cnv_string,
    ((DRAWNS, u'extrusion-shininess'), None): cnv_string,
    ((DRAWNS, u'extrusion-skew'), None): cnv_double,
    ((DRAWNS, u'extrusion-specularity'), None): cnv_string,
    ((DRAWNS, u'extrusion-viewpoint'), None): cnv_string,
    ((DRAWNS, u'fill'), None): cnv_string,
    ((DRAWNS, u'fill-color'), None): cnv_string,
    ((DRAWNS, u'fill-gradient-name'), None): cnv_string,
    ((DRAWNS, u'fill-hatch-name'), None): cnv_string,
    ((DRAWNS, u'fill-hatch-solid'), None): cnv_boolean,
    ((DRAWNS, u'fill-image-height'), None): cnv_lengthorpercent,
    ((DRAWNS, u'fill-image-name'), None): cnv_DrawNameRef,
    ((DRAWNS, u'fill-image-ref-point'), None): cnv_string,
    ((DRAWNS, u'fill-image-ref-point-x'), None): cnv_string,
    ((DRAWNS, u'fill-image-ref-point-y'), None): cnv_string,
    ((DRAWNS, u'fill-image-width'), None): cnv_lengthorpercent,
    ((DRAWNS, u'filter-name'), None): cnv_string,
    ((DRAWNS, u'fit-to-contour'), None): cnv_boolean,
    ((DRAWNS, u'fit-to-size'), None): cnv_string,  # ODF 1.2 says boolean
    ((DRAWNS, u'formula'), None): cnv_string,
    ((DRAWNS, u'frame-display-border'), None): cnv_boolean,
    ((DRAWNS, u'frame-display-scrollbar'), None): cnv_boolean,
    ((DRAWNS, u'frame-margin-horizontal'), None): cnv_string,
    ((DRAWNS, u'frame-margin-vertical'), None): cnv_string,
    ((DRAWNS, u'frame-name'), None): cnv_string,
    ((DRAWNS, u'gamma'), None): cnv_string,
    ((DRAWNS, u'glue-point-leaving-directions'), None): cnv_string,
    ((DRAWNS, u'glue-point-type'), None): cnv_string,
    ((DRAWNS, u'glue-points'), None): cnv_string,
    ((DRAWNS, u'gradient-step-count'), None): cnv_string,
    ((DRAWNS, u'green'), None): cnv_string,
    ((DRAWNS, u'guide-distance'), None): cnv_string,
    ((DRAWNS, u'guide-overhang'), None): cnv_length,
    ((DRAWNS, u'handle-mirror-horizontal'), None): cnv_boolean,
    ((DRAWNS, u'handle-mirror-vertical'), None): cnv_boolean,
    ((DRAWNS, u'handle-polar'), None): cnv_string,
    ((DRAWNS, u'handle-position'), None): cnv_string,
    ((DRAWNS, u'handle-radius-range-maximum'), None): cnv_string,
    ((DRAWNS, u'handle-radius-range-minimum'), None): cnv_string,
    ((DRAWNS, u'handle-range-x-maximum'), None): cnv_string,
    ((DRAWNS, u'handle-range-x-minimum'), None): cnv_string,
    ((DRAWNS, u'handle-range-y-maximum'), None): cnv_string,
    ((DRAWNS, u'handle-range-y-minimum'), None): cnv_string,
    ((DRAWNS, u'handle-switched'), None): cnv_boolean,
    #	((DRAWNS,u'id'), None): cnv_ID,
    #	((DRAWNS,u'id'), None): cnv_nonNegativeInteger,   # ?? line 6581 in RNG
    ((DRAWNS, u'id'), None): cnv_string,
    ((DRAWNS, u'image-opacity'), None): cnv_string,
    ((DRAWNS, u'kind'), None): cnv_string,
    ((DRAWNS, u'layer'), None): cnv_string,
    ((DRAWNS, u'line-distance'), None): cnv_string,
    ((DRAWNS, u'line-skew'), None): cnv_string,
    ((DRAWNS, u'luminance'), None): cnv_string,
    ((DRAWNS, u'marker-end-center'), None): cnv_boolean,
    ((DRAWNS, u'marker-end'), None): cnv_string,
    ((DRAWNS, u'marker-end-width'), None): cnv_length,
    ((DRAWNS, u'marker-start-center'), None): cnv_boolean,
    ((DRAWNS, u'marker-start'), None): cnv_string,
    ((DRAWNS, u'marker-start-width'), None): cnv_length,
    ((DRAWNS, u'master-page-name'), None): cnv_StyleNameRef,
    ((DRAWNS, u'may-script'), None): cnv_boolean,
    ((DRAWNS, u'measure-align'), None): cnv_string,
    ((DRAWNS, u'measure-vertical-align'), None): cnv_string,
    ((DRAWNS, u'mime-type'), None): cnv_string,
    ((DRAWNS, u'mirror-horizontal'), None): cnv_boolean,
    ((DRAWNS, u'mirror-vertical'), None): cnv_boolean,
    ((DRAWNS, u'modifiers'), None): cnv_string,
    ((DRAWNS, u'name'), None): cnv_NCName,
    #	((DRAWNS,u'name'), None): cnv_string,
    ((DRAWNS, u'nav-order'), None): cnv_IDREF,
    ((DRAWNS, u'nohref'), None): cnv_string,
    ((DRAWNS, u'notify-on-update-of-ranges'), None): cnv_string,
    ((DRAWNS, u'object'), None): cnv_string,
    ((DRAWNS, u'ole-draw-aspect'), None): cnv_string,
    ((DRAWNS, u'opacity'), None): cnv_string,
    ((DRAWNS, u'opacity-name'), None): cnv_string,
    ((DRAWNS, u'page-number'), None): cnv_positiveInteger,
    ((DRAWNS, u'parallel'), None): cnv_boolean,
    ((DRAWNS, u'path-stretchpoint-x'), None): cnv_double,
    ((DRAWNS, u'path-stretchpoint-y'), None): cnv_double,
    ((DRAWNS, u'placing'), None): cnv_string,
    ((DRAWNS, u'points'), None): cnv_points,
    ((DRAWNS, u'protected'), None): cnv_boolean,
    ((DRAWNS, u'recreate-on-edit'), None): cnv_boolean,
    ((DRAWNS, u'red'), None): cnv_string,
    ((DRAWNS, u'rotation'), None): cnv_integer,
    ((DRAWNS, u'secondary-fill-color'), None): cnv_string,
    ((DRAWNS, u'shadow'), None): cnv_string,
    ((DRAWNS, u'shadow-color'), None): cnv_string,
    ((DRAWNS, u'shadow-offset-x'), None): cnv_length,
    ((DRAWNS, u'shadow-offset-y'), None): cnv_length,
    ((DRAWNS, u'shadow-opacity'), None): cnv_string,
    ((DRAWNS, u'shape-id'), None): cnv_IDREF,
    ((DRAWNS, u'sharpness'), None): cnv_string,
    ((DRAWNS, u'show-unit'), None): cnv_boolean,
    ((DRAWNS, u'start-angle'), None): cnv_double,
    ((DRAWNS, u'start'), None): cnv_string,
    ((DRAWNS, u'start-color'), None): cnv_string,
    ((DRAWNS, u'start-glue-point'), None): cnv_nonNegativeInteger,
    ((DRAWNS, u'start-guide'), None): cnv_length,
    ((DRAWNS, u'start-intensity'), None): cnv_string,
    ((DRAWNS, u'start-line-spacing-horizontal'), None): cnv_string,
    ((DRAWNS, u'start-line-spacing-vertical'), None): cnv_string,
    ((DRAWNS, u'start-shape'), None): cnv_IDREF,
    ((DRAWNS, u'stroke'), None): cnv_string,
    ((DRAWNS, u'stroke-dash'), None): cnv_string,
    ((DRAWNS, u'stroke-dash-names'), None): cnv_string,
    ((DRAWNS, u'stroke-linejoin'), None): cnv_string,
    ((DRAWNS, u'style'), None): cnv_string,
    ((DRAWNS, u'style-name'), None): cnv_StyleNameRef,
    ((DRAWNS, u'symbol-color'), None): cnv_string,
    ((DRAWNS, u'text-areas'), None): cnv_string,
    ((DRAWNS, u'text-path-allowed'), None): cnv_boolean,
    ((DRAWNS, u'text-path'), None): cnv_boolean,
    ((DRAWNS, u'text-path-mode'), None): cnv_string,
    ((DRAWNS, u'text-path-same-letter-heights'), None): cnv_boolean,
    ((DRAWNS, u'text-path-scale'), None): cnv_string,
    ((DRAWNS, u'text-rotate-angle'), None): cnv_double,
    ((DRAWNS, u'text-style-name'), None): cnv_StyleNameRef,
    ((DRAWNS, u'textarea-horizontal-align'), None): cnv_string,
    ((DRAWNS, u'textarea-vertical-align'), None): cnv_string,
    ((DRAWNS, u'tile-repeat-offset'), None): cnv_string,
    ((DRAWNS, u'transform'), None): cnv_string,
    ((DRAWNS, u'type'), None): cnv_string,
    ((DRAWNS, u'unit'), None): cnv_string,
    ((DRAWNS, u'value'), None): cnv_string,
    ((DRAWNS, u'visible-area-height'), None): cnv_string,
    ((DRAWNS, u'visible-area-left'), None): cnv_string,
    ((DRAWNS, u'visible-area-top'), None): cnv_string,
    ((DRAWNS, u'visible-area-width'), None): cnv_string,
    ((DRAWNS, u'wrap-influence-on-position'), None): cnv_string,
    ((DRAWNS, u'z-index'), None): cnv_nonNegativeInteger,
    ((FONS, u'background-color'), None): cnv_string,
    ((FONS, u'border-bottom'), None): cnv_string,
    ((FONS, u'border'), None): cnv_string,
    ((FONS, u'border-left'), None): cnv_string,
    ((FONS, u'border-right'), None): cnv_string,
    ((FONS, u'border-top'), None): cnv_string,
    ((FONS, u'break-after'), None): cnv_string,
    ((FONS, u'break-before'), None): cnv_string,
    ((FONS, u'clip'), None): cnv_string,
    ((FONS, u'color'), None): cnv_string,
    ((FONS, u'column-count'), None): cnv_positiveInteger,
    ((FONS, u'column-gap'), None): cnv_length,
    ((FONS, u'country'), None): cnv_token,
    ((FONS, u'end-indent'), None): cnv_length,
    ((FONS, u'font-family'), None): cnv_string,
    ((FONS, u'font-size'), None): cnv_string,
    ((FONS, u'font-style'), None): cnv_string,
    ((FONS, u'font-variant'), None): cnv_string,
    ((FONS, u'font-weight'), None): cnv_string,
    ((FONS, u'height'), None): cnv_string,
    ((FONS, u'hyphenate'), None): cnv_boolean,
    ((FONS, u'hyphenation-keep'), None): cnv_string,
    ((FONS, u'hyphenation-ladder-count'), None): cnv_string,
    ((FONS, u'hyphenation-push-char-count'), None): cnv_string,
    ((FONS, u'hyphenation-remain-char-count'), None): cnv_string,
    ((FONS, u'keep-together'), None): cnv_string,
    ((FONS, u'keep-with-next'), None): cnv_string,
    ((FONS, u'language'), None): cnv_token,
    ((FONS, u'letter-spacing'), None): cnv_string,
    ((FONS, u'line-height'), None): cnv_string,
    ((FONS, u'margin-bottom'), None): cnv_string,
    ((FONS, u'margin'), None): cnv_string,
    ((FONS, u'margin-left'), None): cnv_string,
    ((FONS, u'margin-right'), None): cnv_string,
    ((FONS, u'margin-top'), None): cnv_string,
    ((FONS, u'max-height'), None): cnv_string,
    ((FONS, u'max-width'), None): cnv_string,
    ((FONS, u'min-height'), None): cnv_length,
    ((FONS, u'min-width'), None): cnv_string,
    ((FONS, u'orphans'), None): cnv_string,
    ((FONS, u'padding-bottom'), None): cnv_string,
    ((FONS, u'padding'), None): cnv_string,
    ((FONS, u'padding-left'), None): cnv_string,
    ((FONS, u'padding-right'), None): cnv_string,
    ((FONS, u'padding-top'), None): cnv_string,
    ((FONS, u'page-height'), None): cnv_length,
    ((FONS, u'page-width'), None): cnv_length,
    ((FONS, u'space-after'), None): cnv_length,
    ((FONS, u'space-before'), None): cnv_length,
    ((FONS, u'start-indent'), None): cnv_length,
    ((FONS, u'text-align'), None): cnv_string,
    ((FONS, u'text-align-last'), None): cnv_string,
    ((FONS, u'text-indent'), None): cnv_string,
    ((FONS, u'text-shadow'), None): cnv_string,
    ((FONS, u'text-transform'), None): cnv_string,
    ((FONS, u'widows'), None): cnv_string,
    ((FONS, u'width'), None): cnv_string,
    ((FONS, u'wrap-option'), None): cnv_string,
    ((FORMNS, u'allow-deletes'), None): cnv_boolean,
    ((FORMNS, u'allow-inserts'), None): cnv_boolean,
    ((FORMNS, u'allow-updates'), None): cnv_boolean,
    ((FORMNS, u'apply-design-mode'), None): cnv_boolean,
    ((FORMNS, u'apply-filter'), None): cnv_boolean,
    ((FORMNS, u'auto-complete'), None): cnv_boolean,
    ((FORMNS, u'automatic-focus'), None): cnv_boolean,
    ((FORMNS, u'bound-column'), None): cnv_string,
    ((FORMNS, u'button-type'), None): cnv_string,
    ((FORMNS, u'command'), None): cnv_string,
    ((FORMNS, u'command-type'), None): cnv_string,
    ((FORMNS, u'control-implementation'), None): cnv_namespacedToken,
    ((FORMNS, u'convert-empty-to-null'), None): cnv_boolean,
    ((FORMNS, u'current-selected'), None): cnv_boolean,
    ((FORMNS, u'current-state'), None): cnv_string,
    #	((FORMNS,u'current-value'), None): cnv_date,
    #	((FORMNS,u'current-value'), None): cnv_double,
    ((FORMNS, u'current-value'), None): cnv_string,
    #	((FORMNS,u'current-value'), None): cnv_time,
    ((FORMNS, u'data-field'), None): cnv_string,
    ((FORMNS, u'datasource'), None): cnv_string,
    ((FORMNS, u'default-button'), None): cnv_boolean,
    ((FORMNS, u'delay-for-repeat'), None): cnv_duration,
    ((FORMNS, u'detail-fields'), None): cnv_string,
    ((FORMNS, u'disabled'), None): cnv_boolean,
    ((FORMNS, u'dropdown'), None): cnv_boolean,
    ((FORMNS, u'echo-char'), None): cnv_string,
    ((FORMNS, u'enctype'), None): cnv_string,
    ((FORMNS, u'escape-processing'), None): cnv_boolean,
    ((FORMNS, u'filter'), None): cnv_string,
    ((FORMNS, u'focus-on-click'), None): cnv_boolean,
    ((FORMNS, u'for'), None): cnv_string,
    ((FORMNS, u'id'), None): cnv_ID,
    ((FORMNS, u'ignore-result'), None): cnv_boolean,
    ((FORMNS, u'image-align'), None): cnv_string,
    ((FORMNS, u'image-data'), None): cnv_anyURI,
    ((FORMNS, u'image-position'), None): cnv_string,
    ((FORMNS, u'is-tristate'), None): cnv_boolean,
    ((FORMNS, u'label'), None): cnv_string,
    ((FORMNS, u'list-source'), None): cnv_string,
    ((FORMNS, u'list-source-type'), None): cnv_string,
    ((FORMNS, u'master-fields'), None): cnv_string,
    ((FORMNS, u'max-length'), None): cnv_nonNegativeInteger,
    #	((FORMNS,u'max-value'), None): cnv_date,
    #	((FORMNS,u'max-value'), None): cnv_double,
    ((FORMNS, u'max-value'), None): cnv_string,
    #	((FORMNS,u'max-value'), None): cnv_time,
    ((FORMNS, u'method'), None): cnv_string,
    #	((FORMNS,u'min-value'), None): cnv_date,
    #	((FORMNS,u'min-value'), None): cnv_double,
    ((FORMNS, u'min-value'), None): cnv_string,
    #	((FORMNS,u'min-value'), None): cnv_time,
    ((FORMNS, u'multi-line'), None): cnv_boolean,
    ((FORMNS, u'multiple'), None): cnv_boolean,
    ((FORMNS, u'name'), None): cnv_string,
    ((FORMNS, u'navigation-mode'), None): cnv_string,
    ((FORMNS, u'order'), None): cnv_string,
    ((FORMNS, u'orientation'), None): cnv_string,
    ((FORMNS, u'page-step-size'), None): cnv_positiveInteger,
    ((FORMNS, u'printable'), None): cnv_boolean,
    ((FORMNS, u'property-name'), None): cnv_string,
    ((FORMNS, u'readonly'), None): cnv_boolean,
    ((FORMNS, u'selected'), None): cnv_boolean,
    ((FORMNS, u'size'), None): cnv_nonNegativeInteger,
    ((FORMNS, u'state'), None): cnv_string,
    ((FORMNS, u'step-size'), None): cnv_positiveInteger,
    ((FORMNS, u'tab-cycle'), None): cnv_string,
    ((FORMNS, u'tab-index'), None): cnv_nonNegativeInteger,
    ((FORMNS, u'tab-stop'), None): cnv_boolean,
    ((FORMNS, u'text-style-name'), None): cnv_StyleNameRef,
    ((FORMNS, u'title'), None): cnv_string,
    ((FORMNS, u'toggle'), None): cnv_boolean,
    ((FORMNS, u'validation'), None): cnv_boolean,
    #	((FORMNS,u'value'), None): cnv_date,
    #	((FORMNS,u'value'), None): cnv_double,
    ((FORMNS, u'value'), None): cnv_string,
    #	((FORMNS,u'value'), None): cnv_time,
    ((FORMNS, u'visual-effect'), None): cnv_string,
    ((FORMNS, u'xforms-list-source'), None): cnv_string,
    ((FORMNS, u'xforms-submission'), None): cnv_string,
    ((MANIFESTNS, 'algorithm-name'), None): cnv_string,
    ((MANIFESTNS, 'checksum'), None): cnv_string,
    ((MANIFESTNS, 'checksum-type'), None): cnv_string,
    ((MANIFESTNS, 'full-path'), None): cnv_string,
    ((MANIFESTNS, 'initialisation-vector'), None): cnv_string,
    ((MANIFESTNS, 'iteration-count'), None): cnv_nonNegativeInteger,
    ((MANIFESTNS, 'key-derivation-name'), None): cnv_string,
    ((MANIFESTNS, 'media-type'), None): cnv_string,
    ((MANIFESTNS, 'salt'), None): cnv_string,
    ((MANIFESTNS, 'size'), None): cnv_nonNegativeInteger,
    ((METANS, u'cell-count'), None): cnv_nonNegativeInteger,
    ((METANS, u'character-count'), None): cnv_nonNegativeInteger,
    ((METANS, u'date'), None): cnv_dateTime,
    ((METANS, u'delay'), None): cnv_duration,
    ((METANS, u'draw-count'), None): cnv_nonNegativeInteger,
    ((METANS, u'frame-count'), None): cnv_nonNegativeInteger,
    ((METANS, u'image-count'), None): cnv_nonNegativeInteger,
    ((METANS, u'name'), None): cnv_string,
    ((METANS, u'non-whitespace-character-count'), None): cnv_nonNegativeInteger,
    ((METANS, u'object-count'), None): cnv_nonNegativeInteger,
    ((METANS, u'ole-object-count'), None): cnv_nonNegativeInteger,
    ((METANS, u'page-count'), None): cnv_nonNegativeInteger,
    ((METANS, u'paragraph-count'), None): cnv_nonNegativeInteger,
    ((METANS, u'row-count'), None): cnv_nonNegativeInteger,
    ((METANS, u'sentence-count'), None): cnv_nonNegativeInteger,
    ((METANS, u'syllable-count'), None): cnv_nonNegativeInteger,
    ((METANS, u'table-count'), None): cnv_nonNegativeInteger,
    ((METANS, u'value-type'), None): cnv_metavaluetype,
    ((METANS, u'word-count'), None): cnv_nonNegativeInteger,
    ((NUMBERNS, u'automatic-order'), None): cnv_boolean,
    ((NUMBERNS, u'calendar'), None): cnv_string,
    ((NUMBERNS, u'country'), None): cnv_token,
    ((NUMBERNS, u'decimal-places'), None): cnv_integer,
    ((NUMBERNS, u'decimal-replacement'), None): cnv_string,
    ((NUMBERNS, u'denominator-value'), None): cnv_integer,
    ((NUMBERNS, u'display-factor'), None): cnv_double,
    ((NUMBERNS, u'format-source'), None): cnv_string,
    ((NUMBERNS, u'grouping'), None): cnv_boolean,
    ((NUMBERNS, u'language'), None): cnv_token,
    ((NUMBERNS, u'min-denominator-digits'), None): cnv_integer,
    ((NUMBERNS, u'min-exponent-digits'), None): cnv_integer,
    ((NUMBERNS, u'min-integer-digits'), None): cnv_integer,
    ((NUMBERNS, u'min-numerator-digits'), None): cnv_integer,
    ((NUMBERNS, u'position'), None): cnv_integer,
    ((NUMBERNS, u'possessive-form'), None): cnv_boolean,
    ((NUMBERNS, u'style'), None): cnv_string,
    ((NUMBERNS, u'textual'), None): cnv_boolean,
    ((NUMBERNS, u'title'), None): cnv_string,
    ((NUMBERNS, u'transliteration-country'), None): cnv_token,
    ((NUMBERNS, u'transliteration-format'), None): cnv_string,
    ((NUMBERNS, u'transliteration-language'), None): cnv_token,
    ((NUMBERNS, u'transliteration-style'), None): cnv_string,
    ((NUMBERNS, u'truncate-on-overflow'), None): cnv_boolean,
    ((OFFICENS, u'automatic-update'), None): cnv_boolean,
    ((OFFICENS, u'boolean-value'), None): cnv_boolean,
    ((OFFICENS, u'conversion-mode'), None): cnv_string,
    ((OFFICENS, u'currency'), None): cnv_string,
    ((OFFICENS, u'date-value'), None): cnv_dateTime,
    ((OFFICENS, u'dde-application'), None): cnv_string,
    ((OFFICENS, u'dde-item'), None): cnv_string,
    ((OFFICENS, u'dde-topic'), None): cnv_string,
    ((OFFICENS, u'display'), None): cnv_boolean,
    ((OFFICENS, u'mimetype'), None): cnv_string,
    ((OFFICENS, u'name'), None): cnv_string,
    ((OFFICENS, u'process-content'), None): cnv_boolean,
    ((OFFICENS, u'server-map'), None): cnv_boolean,
    ((OFFICENS, u'string-value'), None): cnv_string,
    ((OFFICENS, u'target-frame'), None): cnv_string,
    ((OFFICENS, u'target-frame-name'), None): cnv_string,
    ((OFFICENS, u'time-value'), None): cnv_duration,
    ((OFFICENS, u'title'), None): cnv_string,
    ((OFFICENS, u'value'), None): cnv_double,
    ((OFFICENS, u'value-type'), None): cnv_string,
    ((OFFICENS, u'version'), None): cnv_string,
    ((PRESENTATIONNS, u'action'), None): cnv_string,
    ((PRESENTATIONNS, u'animations'), None): cnv_string,
    ((PRESENTATIONNS, u'background-objects-visible'), None): cnv_boolean,
    ((PRESENTATIONNS, u'background-visible'), None): cnv_boolean,
    ((PRESENTATIONNS, u'class'), None): cnv_string,
    ((PRESENTATIONNS, u'class-names'), None): cnv_NCNames,
    ((PRESENTATIONNS, u'delay'), None): cnv_duration,
    ((PRESENTATIONNS, u'direction'), None): cnv_string,
    ((PRESENTATIONNS, u'display-date-time'), None): cnv_boolean,
    ((PRESENTATIONNS, u'display-footer'), None): cnv_boolean,
    ((PRESENTATIONNS, u'display-header'), None): cnv_boolean,
    ((PRESENTATIONNS, u'display-page-number'), None): cnv_boolean,
    ((PRESENTATIONNS, u'duration'), None): cnv_string,
    ((PRESENTATIONNS, u'effect'), None): cnv_string,
    ((PRESENTATIONNS, u'endless'), None): cnv_boolean,
    ((PRESENTATIONNS, u'force-manual'), None): cnv_boolean,
    ((PRESENTATIONNS, u'full-screen'), None): cnv_boolean,
    ((PRESENTATIONNS, u'group-id'), None): cnv_string,
    ((PRESENTATIONNS, u'master-element'), None): cnv_IDREF,
    ((PRESENTATIONNS, u'mouse-as-pen'), None): cnv_boolean,
    ((PRESENTATIONNS, u'mouse-visible'), None): cnv_boolean,
    ((PRESENTATIONNS, u'name'), None): cnv_string,
    ((PRESENTATIONNS, u'node-type'), None): cnv_string,
    ((PRESENTATIONNS, u'object'), None): cnv_string,
    ((PRESENTATIONNS, u'pages'), None): cnv_string,
    ((PRESENTATIONNS, u'path-id'), None): cnv_string,
    ((PRESENTATIONNS, u'pause'), None): cnv_duration,
    ((PRESENTATIONNS, u'placeholder'), None): cnv_boolean,
    ((PRESENTATIONNS, u'play-full'), None): cnv_boolean,
    ((PRESENTATIONNS, u'presentation-page-layout-name'), None): cnv_StyleNameRef,
    ((PRESENTATIONNS, u'preset-class'), None): cnv_string,
    ((PRESENTATIONNS, u'preset-id'), None): cnv_string,
    ((PRESENTATIONNS, u'preset-sub-type'), None): cnv_string,
    ((PRESENTATIONNS, u'show'), None): cnv_string,
    ((PRESENTATIONNS, u'show-end-of-presentation-slide'), None): cnv_boolean,
    ((PRESENTATIONNS, u'show-logo'), None): cnv_boolean,
    ((PRESENTATIONNS, u'source'), None): cnv_string,
    ((PRESENTATIONNS, u'speed'), None): cnv_string,
    ((PRESENTATIONNS, u'start-page'), None): cnv_string,
    ((PRESENTATIONNS, u'start-scale'), None): cnv_string,
    ((PRESENTATIONNS, u'start-with-navigator'), None): cnv_boolean,
    ((PRESENTATIONNS, u'stay-on-top'), None): cnv_boolean,
    ((PRESENTATIONNS, u'style-name'), None): cnv_StyleNameRef,
    ((PRESENTATIONNS, u'transition-on-click'), None): cnv_string,
    ((PRESENTATIONNS, u'transition-speed'), None): cnv_string,
    ((PRESENTATIONNS, u'transition-style'), None): cnv_string,
    ((PRESENTATIONNS, u'transition-type'), None): cnv_string,
    ((PRESENTATIONNS, u'use-date-time-name'), None): cnv_string,
    ((PRESENTATIONNS, u'use-footer-name'), None): cnv_string,
    ((PRESENTATIONNS, u'use-header-name'), None): cnv_string,
    ((PRESENTATIONNS, u'user-transformed'), None): cnv_boolean,
    ((PRESENTATIONNS, u'verb'), None): cnv_nonNegativeInteger,
    ((PRESENTATIONNS, u'visibility'), None): cnv_string,
    ((SCRIPTNS, u'event-name'), None): cnv_formula,
    ((SCRIPTNS, u'language'), None): cnv_formula,
    ((SCRIPTNS, u'macro-name'), None): cnv_string,
    ((SMILNS, u'accelerate'), None): cnv_double,
    ((SMILNS, u'accumulate'), None): cnv_string,
    ((SMILNS, u'additive'), None): cnv_string,
    ((SMILNS, u'attributeName'), None): cnv_string,
    ((SMILNS, u'autoReverse'), None): cnv_boolean,
    ((SMILNS, u'begin'), None): cnv_string,
    ((SMILNS, u'by'), None): cnv_string,
    ((SMILNS, u'calcMode'), None): cnv_string,
    ((SMILNS, u'decelerate'), None): cnv_double,
    ((SMILNS, u'direction'), None): cnv_string,
    ((SMILNS, u'dur'), None): cnv_string,
    ((SMILNS, u'end'), None): cnv_string,
    ((SMILNS, u'endsync'), None): cnv_string,
    ((SMILNS, u'fadeColor'), None): cnv_string,
    ((SMILNS, u'fill'), None): cnv_string,
    ((SMILNS, u'fillDefault'), None): cnv_string,
    ((SMILNS, u'from'), None): cnv_string,
    ((SMILNS, u'keySplines'), None): cnv_string,
    ((SMILNS, u'keyTimes'), None): cnv_string,
    ((SMILNS, u'mode'), None): cnv_string,
    ((SMILNS, u'repeatCount'), None): cnv_nonNegativeInteger,
    ((SMILNS, u'repeatDur'), None): cnv_string,
    ((SMILNS, u'restart'), None): cnv_string,
    ((SMILNS, u'restartDefault'), None): cnv_string,
    ((SMILNS, u'subtype'), None): cnv_string,
    ((SMILNS, u'targetElement'), None): cnv_IDREF,
    ((SMILNS, u'to'), None): cnv_string,
    ((SMILNS, u'type'), None): cnv_string,
    ((SMILNS, u'values'), None): cnv_string,
    ((STYLENS, u'adjustment'), None): cnv_string,
    ((STYLENS, u'apply-style-name'), None): cnv_StyleNameRef,
    ((STYLENS, u'auto-text-indent'), None): cnv_boolean,
    ((STYLENS, u'auto-update'), None): cnv_boolean,
    ((STYLENS, u'background-transparency'), None): cnv_string,
    ((STYLENS, u'base-cell-address'), None): cnv_string,
    ((STYLENS, u'border-line-width-bottom'), None): cnv_string,
    ((STYLENS, u'border-line-width'), None): cnv_string,
    ((STYLENS, u'border-line-width-left'), None): cnv_string,
    ((STYLENS, u'border-line-width-right'), None): cnv_string,
    ((STYLENS, u'border-line-width-top'), None): cnv_string,
    ((STYLENS, u'cell-protect'), None): cnv_string,
    ((STYLENS, u'char'), None): cnv_string,
    ((STYLENS, u'class'), None): cnv_string,
    ((STYLENS, u'color'), None): cnv_string,
    ((STYLENS, u'column-width'), None): cnv_string,
    ((STYLENS, u'condition'), None): cnv_string,
    ((STYLENS, u'country-asian'), None): cnv_string,
    ((STYLENS, u'country-complex'), None): cnv_string,
    ((STYLENS, u'data-style-name'), None): cnv_StyleNameRef,
    ((STYLENS, u'decimal-places'), None): cnv_string,
    ((STYLENS, u'default-outline-level'), None): cnv_positiveInteger,
    ((STYLENS, u'diagonal-bl-tr'), None): cnv_string,
    ((STYLENS, u'diagonal-bl-tr-widths'), None): cnv_string,
    ((STYLENS, u'diagonal-tl-br'), None): cnv_string,
    ((STYLENS, u'diagonal-tl-br-widths'), None): cnv_string,
    ((STYLENS, u'direction'), None): cnv_string,
    ((STYLENS, u'display'), None): cnv_boolean,
    ((STYLENS, u'display-name'), None): cnv_string,
    ((STYLENS, u'distance-after-sep'), None): cnv_length,
    ((STYLENS, u'distance-before-sep'), None): cnv_length,
    ((STYLENS, u'distance'), None): cnv_length,
    ((STYLENS, u'dynamic-spacing'), None): cnv_boolean,
    ((STYLENS, u'editable'), None): cnv_boolean,
    ((STYLENS, u'family'), None): cnv_family,
    ((STYLENS, u'filter-name'), None): cnv_string,
    ((STYLENS, u'first-page-number'), None): cnv_string,
    ((STYLENS, u'flow-with-text'), None): cnv_boolean,
    ((STYLENS, u'font-adornments'), None): cnv_string,
    ((STYLENS, u'font-charset'), None): cnv_string,
    ((STYLENS, u'font-charset-asian'), None): cnv_string,
    ((STYLENS, u'font-charset-complex'), None): cnv_string,
    ((STYLENS, u'font-family-asian'), None): cnv_string,
    ((STYLENS, u'font-family-complex'), None): cnv_string,
    ((STYLENS, u'font-family-generic-asian'), None): cnv_string,
    ((STYLENS, u'font-family-generic'), None): cnv_string,
    ((STYLENS, u'font-family-generic-complex'), None): cnv_string,
    ((STYLENS, u'font-independent-line-spacing'), None): cnv_boolean,
    ((STYLENS, u'font-name-asian'), None): cnv_string,
    ((STYLENS, u'font-name'), None): cnv_string,
    ((STYLENS, u'font-name-complex'), None): cnv_string,
    ((STYLENS, u'font-pitch-asian'), None): cnv_string,
    ((STYLENS, u'font-pitch'), None): cnv_string,
    ((STYLENS, u'font-pitch-complex'), None): cnv_string,
    ((STYLENS, u'font-relief'), None): cnv_string,
    ((STYLENS, u'font-size-asian'), None): cnv_string,
    ((STYLENS, u'font-size-complex'), None): cnv_string,
    ((STYLENS, u'font-size-rel-asian'), None): cnv_length,
    ((STYLENS, u'font-size-rel'), None): cnv_length,
    ((STYLENS, u'font-size-rel-complex'), None): cnv_length,
    ((STYLENS, u'font-style-asian'), None): cnv_string,
    ((STYLENS, u'font-style-complex'), None): cnv_string,
    ((STYLENS, u'font-style-name-asian'), None): cnv_string,
    ((STYLENS, u'font-style-name'), None): cnv_string,
    ((STYLENS, u'font-style-name-complex'), None): cnv_string,
    ((STYLENS, u'font-weight-asian'), None): cnv_string,
    ((STYLENS, u'font-weight-complex'), None): cnv_string,
    ((STYLENS, u'footnote-max-height'), None): cnv_length,
    ((STYLENS, u'glyph-orientation-vertical'), None): cnv_string,
    ((STYLENS, u'height'), None): cnv_string,
    ((STYLENS, u'horizontal-pos'), None): cnv_string,
    ((STYLENS, u'horizontal-rel'), None): cnv_string,
    ((STYLENS, u'justify-single-word'), None): cnv_boolean,
    ((STYLENS, u'language-asian'), None): cnv_string,
    ((STYLENS, u'language-complex'), None): cnv_string,
    ((STYLENS, u'layout-grid-base-height'), None): cnv_length,
    ((STYLENS, u'layout-grid-color'), None): cnv_string,
    ((STYLENS, u'layout-grid-display'), None): cnv_boolean,
    ((STYLENS, u'layout-grid-lines'), None): cnv_string,
    ((STYLENS, u'layout-grid-mode'), None): cnv_string,
    ((STYLENS, u'layout-grid-print'), None): cnv_boolean,
    ((STYLENS, u'layout-grid-ruby-below'), None): cnv_boolean,
    ((STYLENS, u'layout-grid-ruby-height'), None): cnv_length,
    ((STYLENS, u'leader-char'), None): cnv_string,
    ((STYLENS, u'leader-color'), None): cnv_string,
    ((STYLENS, u'leader-style'), None): cnv_string,
    ((STYLENS, u'leader-text'), None): cnv_string,
    ((STYLENS, u'leader-text-style'), None): cnv_StyleNameRef,
    ((STYLENS, u'leader-type'), None): cnv_string,
    ((STYLENS, u'leader-width'), None): cnv_string,
    ((STYLENS, u'legend-expansion-aspect-ratio'), None): cnv_double,
    ((STYLENS, u'legend-expansion'), None): cnv_string,
    ((STYLENS, u'length'), None): cnv_positiveInteger,
    ((STYLENS, u'letter-kerning'), None): cnv_boolean,
    ((STYLENS, u'line-break'), None): cnv_string,
    ((STYLENS, u'line-height-at-least'), None): cnv_string,
    ((STYLENS, u'line-spacing'), None): cnv_length,
    ((STYLENS, u'line-style'), None): cnv_string,
    ((STYLENS, u'lines'), None): cnv_positiveInteger,
    ((STYLENS, u'list-style-name'), None): cnv_StyleNameRef,
    ((STYLENS, u'master-page-name'), None): cnv_StyleNameRef,
    ((STYLENS, u'may-break-between-rows'), None): cnv_boolean,
    ((STYLENS, u'min-row-height'), None): cnv_string,
    ((STYLENS, u'mirror'), None): cnv_string,
    ((STYLENS, u'name'), None): cnv_NCName,
    ((STYLENS, u'name'), (STYLENS, u'font-face')): cnv_string,
    ((STYLENS, u'next-style-name'), None): cnv_StyleNameRef,
    ((STYLENS, u'num-format'), None): cnv_string,
    ((STYLENS, u'num-letter-sync'), None): cnv_boolean,
    ((STYLENS, u'num-prefix'), None): cnv_string,
    ((STYLENS, u'num-suffix'), None): cnv_string,
    ((STYLENS, u'number-wrapped-paragraphs'), None): cnv_string,
    ((STYLENS, u'overflow-behavior'), None): cnv_string,
    ((STYLENS, u'page-layout-name'), None): cnv_StyleNameRef,
    ((STYLENS, u'page-number'), None): cnv_string,
    ((STYLENS, u'page-usage'), None): cnv_string,
    ((STYLENS, u'paper-tray-name'), None): cnv_string,
    ((STYLENS, u'parent-style-name'), None): cnv_StyleNameRef,
    ((STYLENS, u'position'), (STYLENS, u'tab-stop')): cnv_length,
    ((STYLENS, u'position'), None): cnv_string,
    ((STYLENS, u'print'), None): cnv_string,
    ((STYLENS, u'print-content'), None): cnv_boolean,
    ((STYLENS, u'print-orientation'), None): cnv_string,
    ((STYLENS, u'print-page-order'), None): cnv_string,
    ((STYLENS, u'protect'), (STYLENS, u'section-properties')): cnv_boolean,
    ((STYLENS, u'protect'), (STYLENS, u'graphic-properties')): cnv_string,
    #	((STYLENS,u'protect'), None): cnv_boolean,
    ((STYLENS, u'punctuation-wrap'), None): cnv_string,
    ((STYLENS, u'register-true'), None): cnv_boolean,
    ((STYLENS, u'register-truth-ref-style-name'), None): cnv_string,
    ((STYLENS, u'rel-column-width'), None): cnv_string,
    ((STYLENS, u'rel-height'), None): cnv_string,
    ((STYLENS, u'rel-width'), None): cnv_string,
    ((STYLENS, u'repeat'), None): cnv_string,
    ((STYLENS, u'repeat-content'), None): cnv_boolean,
    ((STYLENS, u'rotation-align'), None): cnv_string,
    ((STYLENS, u'rotation-angle'), None): cnv_string,
    ((STYLENS, u'row-height'), None): cnv_string,
    ((STYLENS, u'ruby-align'), None): cnv_string,
    ((STYLENS, u'ruby-position'), None): cnv_string,
    ((STYLENS, u'run-through'), None): cnv_string,
    ((STYLENS, u'scale-to'), None): cnv_string,
    ((STYLENS, u'scale-to-pages'), None): cnv_string,
    ((STYLENS, u'script-type'), None): cnv_string,
    ((STYLENS, u'shadow'), None): cnv_string,
    ((STYLENS, u'shrink-to-fit'), None): cnv_boolean,
    ((STYLENS, u'snap-to-layout-grid'), None): cnv_boolean,
    ((STYLENS, u'style'), None): cnv_string,
    ((STYLENS, u'style-name'), None): cnv_StyleNameRef,
    ((STYLENS, u'tab-stop-distance'), None): cnv_string,
    ((STYLENS, u'table-centering'), None): cnv_string,
    ((STYLENS, u'text-align-source'), None): cnv_string,
    ((STYLENS, u'text-autospace'), None): cnv_string,
    ((STYLENS, u'text-blinking'), None): cnv_boolean,
    ((STYLENS, u'text-combine'), None): cnv_string,
    ((STYLENS, u'text-combine-end-char'), None): cnv_string,
    ((STYLENS, u'text-combine-start-char'), None): cnv_string,
    ((STYLENS, u'text-emphasize'), None): cnv_string,
    ((STYLENS, u'text-line-through-color'), None): cnv_string,
    ((STYLENS, u'text-line-through-mode'), None): cnv_string,
    ((STYLENS, u'text-line-through-style'), None): cnv_string,
    ((STYLENS, u'text-line-through-text'), None): cnv_string,
    ((STYLENS, u'text-line-through-text-style'), None): cnv_string,
    ((STYLENS, u'text-line-through-type'), None): cnv_string,
    ((STYLENS, u'text-line-through-width'), None): cnv_string,
    ((STYLENS, u'text-outline'), None): cnv_boolean,
    ((STYLENS, u'text-position'), None): cnv_string,
    ((STYLENS, u'text-rotation-angle'), None): cnv_string,
    ((STYLENS, u'text-rotation-scale'), None): cnv_string,
    ((STYLENS, u'text-scale'), None): cnv_string,
    ((STYLENS, u'text-underline-color'), None): cnv_string,
    ((STYLENS, u'text-underline-mode'), None): cnv_string,
    ((STYLENS, u'text-underline-style'), None): cnv_string,
    ((STYLENS, u'text-underline-type'), None): cnv_string,
    ((STYLENS, u'text-underline-width'), None): cnv_string,
    ((STYLENS, u'type'), None): cnv_string,
    ((STYLENS, u'use-optimal-column-width'), None): cnv_boolean,
    ((STYLENS, u'use-optimal-row-height'), None): cnv_boolean,
    ((STYLENS, u'use-window-font-color'), None): cnv_boolean,
    ((STYLENS, u'vertical-align'), None): cnv_string,
    ((STYLENS, u'vertical-pos'), None): cnv_string,
    ((STYLENS, u'vertical-rel'), None): cnv_string,
    ((STYLENS, u'volatile'), None): cnv_boolean,
    ((STYLENS, u'width'), None): cnv_string,
    ((STYLENS, u'wrap'), None): cnv_string,
    ((STYLENS, u'wrap-contour'), None): cnv_boolean,
    ((STYLENS, u'wrap-contour-mode'), None): cnv_string,
    ((STYLENS, u'wrap-dynamic-threshold'), None): cnv_length,
    ((STYLENS, u'writing-mode-automatic'), None): cnv_boolean,
    ((STYLENS, u'writing-mode'), None): cnv_string,
    ((SVGNS, u'accent-height'), None): cnv_integer,
    ((SVGNS, u'alphabetic'), None): cnv_integer,
    ((SVGNS, u'ascent'), None): cnv_integer,
    ((SVGNS, u'bbox'), None): cnv_string,
    ((SVGNS, u'cap-height'), None): cnv_integer,
    ((SVGNS, u'cx'), None): cnv_string,
    ((SVGNS, u'cy'), None): cnv_string,
    ((SVGNS, u'd'), None): cnv_string,
    ((SVGNS, u'descent'), None): cnv_integer,
    ((SVGNS, u'fill-rule'), None): cnv_string,
    ((SVGNS, u'font-family'), None): cnv_string,
    ((SVGNS, u'font-size'), None): cnv_string,
    ((SVGNS, u'font-stretch'), None): cnv_string,
    ((SVGNS, u'font-style'), None): cnv_string,
    ((SVGNS, u'font-variant'), None): cnv_string,
    ((SVGNS, u'font-weight'), None): cnv_string,
    ((SVGNS, u'fx'), None): cnv_string,
    ((SVGNS, u'fy'), None): cnv_string,
    ((SVGNS, u'gradientTransform'), None): cnv_string,
    ((SVGNS, u'gradientUnits'), None): cnv_string,
    ((SVGNS, u'hanging'), None): cnv_integer,
    ((SVGNS, u'height'), None): cnv_length,
    ((SVGNS, u'ideographic'), None): cnv_integer,
    ((SVGNS, u'mathematical'), None): cnv_integer,
    ((SVGNS, u'name'), None): cnv_string,
    ((SVGNS, u'offset'), None): cnv_string,
    ((SVGNS, u'origin'), None): cnv_string,
    ((SVGNS, u'overline-position'), None): cnv_integer,
    ((SVGNS, u'overline-thickness'), None): cnv_integer,
    ((SVGNS, u'panose-1'), None): cnv_string,
    ((SVGNS, u'path'), None): cnv_string,
    ((SVGNS, u'r'), None): cnv_length,
    ((SVGNS, u'rx'), None): cnv_length,
    ((SVGNS, u'ry'), None): cnv_length,
    ((SVGNS, u'slope'), None): cnv_integer,
    ((SVGNS, u'spreadMethod'), None): cnv_string,
    ((SVGNS, u'stemh'), None): cnv_integer,
    ((SVGNS, u'stemv'), None): cnv_integer,
    ((SVGNS, u'stop-color'), None): cnv_string,
    ((SVGNS, u'stop-opacity'), None): cnv_double,
    ((SVGNS, u'strikethrough-position'), None): cnv_integer,
    ((SVGNS, u'strikethrough-thickness'), None): cnv_integer,
    ((SVGNS, u'string'), None): cnv_string,
    ((SVGNS, u'stroke-color'), None): cnv_string,
    ((SVGNS, u'stroke-opacity'), None): cnv_string,
    ((SVGNS, u'stroke-width'), None): cnv_length,
    ((SVGNS, u'type'), None): cnv_string,
    ((SVGNS, u'underline-position'), None): cnv_integer,
    ((SVGNS, u'underline-thickness'), None): cnv_integer,
    ((SVGNS, u'unicode-range'), None): cnv_string,
    ((SVGNS, u'units-per-em'), None): cnv_integer,
    ((SVGNS, u'v-alphabetic'), None): cnv_integer,
    ((SVGNS, u'v-hanging'), None): cnv_integer,
    ((SVGNS, u'v-ideographic'), None): cnv_integer,
    ((SVGNS, u'v-mathematical'), None): cnv_integer,
    ((SVGNS, u'viewBox'), None): cnv_viewbox,
    ((SVGNS, u'width'), None): cnv_length,
    ((SVGNS, u'widths'), None): cnv_string,
    ((SVGNS, u'x'), None): cnv_length,
    ((SVGNS, u'x-height'), None): cnv_integer,
    ((SVGNS, u'x1'), None): cnv_lengthorpercent,
    ((SVGNS, u'x2'), None): cnv_lengthorpercent,
    ((SVGNS, u'y'), None): cnv_length,
    ((SVGNS, u'y1'), None): cnv_lengthorpercent,
    ((SVGNS, u'y2'), None): cnv_lengthorpercent,
    ((TABLENS, u'acceptance-state'), None): cnv_string,
    ((TABLENS, u'add-empty-lines'), None): cnv_boolean,
    ((TABLENS, u'algorithm'), None): cnv_formula,
    ((TABLENS, u'align'), None): cnv_string,
    ((TABLENS, u'allow-empty-cell'), None): cnv_boolean,
    ((TABLENS, u'application-data'), None): cnv_string,
    ((TABLENS, u'automatic-find-labels'), None): cnv_boolean,
    ((TABLENS, u'base-cell-address'), None): cnv_string,
    ((TABLENS, u'bind-styles-to-content'), None): cnv_boolean,
    ((TABLENS, u'border-color'), None): cnv_string,
    ((TABLENS, u'border-model'), None): cnv_string,
    ((TABLENS, u'buttons'), None): cnv_string,
    ((TABLENS, u'buttons'), None): cnv_string,
    ((TABLENS, u'case-sensitive'), None): cnv_boolean,
    ((TABLENS, u'case-sensitive'), None): cnv_string,
    ((TABLENS, u'cell-address'), None): cnv_string,
    ((TABLENS, u'cell-range-address'), None): cnv_string,
    ((TABLENS, u'cell-range-address'), None): cnv_string,
    ((TABLENS, u'cell-range'), None): cnv_string,
    ((TABLENS, u'column'), None): cnv_integer,
    ((TABLENS, u'comment'), None): cnv_string,
    ((TABLENS, u'condition'), None): cnv_formula,
    ((TABLENS, u'condition-source'), None): cnv_string,
    ((TABLENS, u'condition-source-range-address'), None): cnv_string,
    ((TABLENS, u'contains-error'), None): cnv_boolean,
    ((TABLENS, u'contains-header'), None): cnv_boolean,
    ((TABLENS, u'content-validation-name'), None): cnv_string,
    ((TABLENS, u'copy-back'), None): cnv_boolean,
    ((TABLENS, u'copy-formulas'), None): cnv_boolean,
    ((TABLENS, u'copy-styles'), None): cnv_boolean,
    ((TABLENS, u'count'), None): cnv_positiveInteger,
    ((TABLENS, u'country'), None): cnv_token,
    ((TABLENS, u'data-cell-range-address'), None): cnv_string,
    ((TABLENS, u'data-field'), None): cnv_string,
    ((TABLENS, u'data-type'), None): cnv_string,
    ((TABLENS, u'database-name'), None): cnv_string,
    ((TABLENS, u'database-table-name'), None): cnv_string,
    ((TABLENS, u'date-end'), None): cnv_string,
    ((TABLENS, u'date-start'), None): cnv_string,
    ((TABLENS, u'date-value'), None): cnv_date,
    ((TABLENS, u'default-cell-style-name'), None): cnv_StyleNameRef,
    ((TABLENS, u'direction'), None): cnv_string,
    ((TABLENS, u'display-border'), None): cnv_boolean,
    ((TABLENS, u'display'), None): cnv_boolean,
    ((TABLENS, u'display-duplicates'), None): cnv_boolean,
    ((TABLENS, u'display-filter-buttons'), None): cnv_boolean,
    ((TABLENS, u'display-list'), None): cnv_string,
    ((TABLENS, u'display-member-mode'), None): cnv_string,
    ((TABLENS, u'drill-down-on-double-click'), None): cnv_boolean,
    ((TABLENS, u'enabled'), None): cnv_boolean,
    ((TABLENS, u'end-cell-address'), None): cnv_string,
    ((TABLENS, u'end'), None): cnv_string,
    ((TABLENS, u'end-column'), None): cnv_integer,
    ((TABLENS, u'end-position'), None): cnv_integer,
    ((TABLENS, u'end-row'), None): cnv_integer,
    ((TABLENS, u'end-table'), None): cnv_integer,
    ((TABLENS, u'end-x'), None): cnv_length,
    ((TABLENS, u'end-y'), None): cnv_length,
    ((TABLENS, u'execute'), None): cnv_boolean,
    ((TABLENS, u'expression'), None): cnv_formula,
    ((TABLENS, u'field-name'), None): cnv_string,
    ((TABLENS, u'field-number'), None): cnv_nonNegativeInteger,
    ((TABLENS, u'field-number'), None): cnv_string,
    ((TABLENS, u'filter-name'), None): cnv_string,
    ((TABLENS, u'filter-options'), None): cnv_string,
    ((TABLENS, u'formula'), None): cnv_formula,
    ((TABLENS, u'function'), None): cnv_string,
    ((TABLENS, u'function'), None): cnv_string,
    ((TABLENS, u'grand-total'), None): cnv_string,
    ((TABLENS, u'group-by-field-number'), None): cnv_nonNegativeInteger,
    ((TABLENS, u'grouped-by'), None): cnv_string,
    ((TABLENS, u'has-persistent-data'), None): cnv_boolean,
    ((TABLENS, u'id'), None): cnv_string,
    ((TABLENS, u'identify-categories'), None): cnv_boolean,
    ((TABLENS, u'ignore-empty-rows'), None): cnv_boolean,
    ((TABLENS, u'index'), None): cnv_nonNegativeInteger,
    ((TABLENS, u'is-active'), None): cnv_boolean,
    ((TABLENS, u'is-data-layout-field'), None): cnv_string,
    ((TABLENS, u'is-selection'), None): cnv_boolean,
    ((TABLENS, u'is-sub-table'), None): cnv_boolean,
    ((TABLENS, u'label-cell-range-address'), None): cnv_string,
    ((TABLENS, u'language'), None): cnv_token,
    ((TABLENS, u'language'), None): cnv_token,
    ((TABLENS, u'last-column-spanned'), None): cnv_positiveInteger,
    ((TABLENS, u'last-row-spanned'), None): cnv_positiveInteger,
    ((TABLENS, u'layout-mode'), None): cnv_string,
    ((TABLENS, u'link-to-source-data'), None): cnv_boolean,
    ((TABLENS, u'marked-invalid'), None): cnv_boolean,
    ((TABLENS, u'matrix-covered'), None): cnv_boolean,
    ((TABLENS, u'maximum-difference'), None): cnv_double,
    ((TABLENS, u'member-count'), None): cnv_nonNegativeInteger,
    ((TABLENS, u'member-name'), None): cnv_string,
    ((TABLENS, u'member-type'), None): cnv_string,
    ((TABLENS, u'message-type'), None): cnv_string,
    ((TABLENS, u'mode'), None): cnv_string,
    ((TABLENS, u'multi-deletion-spanned'), None): cnv_integer,
    ((TABLENS, u'name'), None): cnv_string,
    ((TABLENS, u'name'), None): cnv_string,
    ((TABLENS, u'null-year'), None): cnv_positiveInteger,
    ((TABLENS, u'number-columns-repeated'), None): cnv_positiveInteger,
    ((TABLENS, u'number-columns-spanned'), None): cnv_positiveInteger,
    ((TABLENS, u'number-matrix-columns-spanned'), None): cnv_positiveInteger,
    ((TABLENS, u'number-matrix-rows-spanned'), None): cnv_positiveInteger,
    ((TABLENS, u'number-rows-repeated'), None): cnv_positiveInteger,
    ((TABLENS, u'number-rows-spanned'), None): cnv_positiveInteger,
    ((TABLENS, u'object-name'), None): cnv_string,
    ((TABLENS, u'on-update-keep-size'), None): cnv_boolean,
    ((TABLENS, u'on-update-keep-styles'), None): cnv_boolean,
    ((TABLENS, u'operator'), None): cnv_string,
    ((TABLENS, u'operator'), None): cnv_string,
    ((TABLENS, u'order'), None): cnv_string,
    ((TABLENS, u'orientation'), None): cnv_string,
    ((TABLENS, u'orientation'), None): cnv_string,
    ((TABLENS, u'page-breaks-on-group-change'), None): cnv_boolean,
    ((TABLENS, u'parse-sql-statement'), None): cnv_boolean,
    ((TABLENS, u'password'), None): cnv_string,
    ((TABLENS, u'position'), None): cnv_integer,
    ((TABLENS, u'precision-as-shown'), None): cnv_boolean,
    ((TABLENS, u'print'), None): cnv_boolean,
    ((TABLENS, u'print-ranges'), None): cnv_string,
    ((TABLENS, u'protect'), None): cnv_boolean,
    ((TABLENS, u'protected'), None): cnv_boolean,
    ((TABLENS, u'protection-key'), None): cnv_string,
    ((TABLENS, u'query-name'), None): cnv_string,
    ((TABLENS, u'range-usable-as'), None): cnv_string,
    ((TABLENS, u'refresh-delay'), None): cnv_boolean,
    ((TABLENS, u'refresh-delay'), None): cnv_duration,
    ((TABLENS, u'rejecting-change-id'), None): cnv_string,
    ((TABLENS, u'row'), None): cnv_integer,
    ((TABLENS, u'scenario-ranges'), None): cnv_string,
    ((TABLENS, u'search-criteria-must-apply-to-whole-cell'), None): cnv_boolean,
    ((TABLENS, u'selected-page'), None): cnv_string,
    ((TABLENS, u'show-details'), None): cnv_boolean,
    ((TABLENS, u'show-empty'), None): cnv_boolean,
    ((TABLENS, u'show-empty'), None): cnv_string,
    ((TABLENS, u'show-filter-button'), None): cnv_boolean,
    ((TABLENS, u'sort-mode'), None): cnv_string,
    ((TABLENS, u'source-cell-range-addresses'), None): cnv_string,
    ((TABLENS, u'source-cell-range-addresses'), None): cnv_string,
    ((TABLENS, u'source-field-name'), None): cnv_string,
    ((TABLENS, u'source-field-name'), None): cnv_string,
    ((TABLENS, u'source-name'), None): cnv_string,
    ((TABLENS, u'sql-statement'), None): cnv_string,
    ((TABLENS, u'start'), None): cnv_string,
    ((TABLENS, u'start-column'), None): cnv_integer,
    ((TABLENS, u'start-position'), None): cnv_integer,
    ((TABLENS, u'start-row'), None): cnv_integer,
    ((TABLENS, u'start-table'), None): cnv_integer,
    ((TABLENS, u'status'), None): cnv_string,
    ((TABLENS, u'step'), None): cnv_double,
    ((TABLENS, u'steps'), None): cnv_positiveInteger,
    ((TABLENS, u'structure-protected'), None): cnv_boolean,
    ((TABLENS, u'style-name'), None): cnv_StyleNameRef,
    ((TABLENS, u'table-background'), None): cnv_boolean,
    ((TABLENS, u'table'), None): cnv_integer,
    ((TABLENS, u'table-name'), None): cnv_string,
    ((TABLENS, u'target-cell-address'), None): cnv_string,
    ((TABLENS, u'target-cell-address'), None): cnv_string,
    ((TABLENS, u'target-range-address'), None): cnv_string,
    ((TABLENS, u'target-range-address'), None): cnv_string,
    ((TABLENS, u'title'), None): cnv_string,
    ((TABLENS, u'track-changes'), None): cnv_boolean,
    ((TABLENS, u'type'), None): cnv_string,
    ((TABLENS, u'use-labels'), None): cnv_string,
    ((TABLENS, u'use-regular-expressions'), None): cnv_boolean,
    ((TABLENS, u'used-hierarchy'), None): cnv_integer,
    ((TABLENS, u'user-name'), None): cnv_string,
    ((TABLENS, u'value'), None): cnv_string,
    ((TABLENS, u'value'), None): cnv_string,
    ((TABLENS, u'value-type'), None): cnv_string,
    ((TABLENS, u'visibility'), None): cnv_string,
    ((TEXTNS, u'active'), None): cnv_boolean,
    ((TEXTNS, u'address'), None): cnv_string,
    ((TEXTNS, u'alphabetical-separators'), None): cnv_boolean,
    ((TEXTNS, u'anchor-page-number'), None): cnv_positiveInteger,
    ((TEXTNS, u'anchor-type'), None): cnv_string,
    ((TEXTNS, u'animation'), None): cnv_string,
    ((TEXTNS, u'animation-delay'), None): cnv_string,
    ((TEXTNS, u'animation-direction'), None): cnv_string,
    ((TEXTNS, u'animation-repeat'), None): cnv_string,
    ((TEXTNS, u'animation-start-inside'), None): cnv_boolean,
    ((TEXTNS, u'animation-steps'), None): cnv_length,
    ((TEXTNS, u'animation-stop-inside'), None): cnv_boolean,
    ((TEXTNS, u'annote'), None): cnv_string,
    ((TEXTNS, u'author'), None): cnv_string,
    ((TEXTNS, u'bibliography-data-field'), None): cnv_string,
    ((TEXTNS, u'bibliography-type'), None): cnv_string,
    ((TEXTNS, u'booktitle'), None): cnv_string,
    ((TEXTNS, u'bullet-char'), None): cnv_string,
    ((TEXTNS, u'bullet-relative-size'), None): cnv_string,
    ((TEXTNS, u'c'), None): cnv_nonNegativeInteger,
    ((TEXTNS, u'capitalize-entries'), None): cnv_boolean,
    ((TEXTNS, u'caption-sequence-format'), None): cnv_string,
    ((TEXTNS, u'caption-sequence-name'), None): cnv_string,
    ((TEXTNS, u'change-id'), None): cnv_IDREF,
    ((TEXTNS, u'chapter'), None): cnv_string,
    ((TEXTNS, u'citation-body-style-name'), None): cnv_StyleNameRef,
    ((TEXTNS, u'citation-style-name'), None): cnv_StyleNameRef,
    ((TEXTNS, u'class-names'), None): cnv_NCNames,
    ((TEXTNS, u'column-name'), None): cnv_string,
    ((TEXTNS, u'combine-entries'), None): cnv_boolean,
    ((TEXTNS, u'combine-entries-with-dash'), None): cnv_boolean,
    ((TEXTNS, u'combine-entries-with-pp'), None): cnv_boolean,
    ((TEXTNS, u'comma-separated'), None): cnv_boolean,
    ((TEXTNS, u'cond-style-name'), None): cnv_StyleNameRef,
    ((TEXTNS, u'condition'), None): cnv_formula,
    ((TEXTNS, u'connection-name'), None): cnv_string,
    ((TEXTNS, u'consecutive-numbering'), None): cnv_boolean,
    ((TEXTNS, u'continue-numbering'), None): cnv_boolean,
    ((TEXTNS, u'copy-outline-levels'), None): cnv_boolean,
    ((TEXTNS, u'count-empty-lines'), None): cnv_boolean,
    ((TEXTNS, u'count-in-text-boxes'), None): cnv_boolean,
    ((TEXTNS, u'current-value'), None): cnv_boolean,
    ((TEXTNS, u'custom1'), None): cnv_string,
    ((TEXTNS, u'custom2'), None): cnv_string,
    ((TEXTNS, u'custom3'), None): cnv_string,
    ((TEXTNS, u'custom4'), None): cnv_string,
    ((TEXTNS, u'custom5'), None): cnv_string,
    ((TEXTNS, u'database-name'), None): cnv_string,
    ((TEXTNS, u'date-adjust'), None): cnv_duration,
    ((TEXTNS, u'date-value'), None): cnv_date,
    #	((TEXTNS,u'date-value'), None): cnv_dateTime,
    ((TEXTNS, u'default-style-name'), None): cnv_StyleNameRef,
    ((TEXTNS, u'description'), None): cnv_string,
    ((TEXTNS, u'display'), None): cnv_string,
    ((TEXTNS, u'display-levels'), None): cnv_positiveInteger,
    ((TEXTNS, u'display-outline-level'), None): cnv_nonNegativeInteger,
    ((TEXTNS, u'dont-balance-text-columns'), None): cnv_boolean,
    ((TEXTNS, u'duration'), None): cnv_duration,
    ((TEXTNS, u'edition'), None): cnv_string,
    ((TEXTNS, u'editor'), None): cnv_string,
    ((TEXTNS, u'filter-name'), None): cnv_string,
    ((TEXTNS, u'first-row-end-column'), None): cnv_string,
    ((TEXTNS, u'first-row-start-column'), None): cnv_string,
    ((TEXTNS, u'fixed'), None): cnv_boolean,
    ((TEXTNS, u'footnotes-position'), None): cnv_string,
    ((TEXTNS, u'formula'), None): cnv_formula,
    ((TEXTNS, u'global'), None): cnv_boolean,
    ((TEXTNS, u'howpublished'), None): cnv_string,
    ((TEXTNS, u'id'), None): cnv_ID,
    #	((TEXTNS,u'id'), None): cnv_string,
    ((TEXTNS, u'identifier'), None): cnv_string,
    ((TEXTNS, u'ignore-case'), None): cnv_boolean,
    ((TEXTNS, u'increment'), None): cnv_nonNegativeInteger,
    ((TEXTNS, u'index-name'), None): cnv_string,
    ((TEXTNS, u'index-scope'), None): cnv_string,
    ((TEXTNS, u'institution'), None): cnv_string,
    ((TEXTNS, u'is-hidden'), None): cnv_boolean,
    ((TEXTNS, u'is-list-header'), None): cnv_boolean,
    ((TEXTNS, u'isbn'), None): cnv_string,
    ((TEXTNS, u'issn'), None): cnv_string,
    ((TEXTNS, u'issn'), None): cnv_string,
    ((TEXTNS, u'journal'), None): cnv_string,
    ((TEXTNS, u'key'), None): cnv_string,
    ((TEXTNS, u'key1'), None): cnv_string,
    ((TEXTNS, u'key1-phonetic'), None): cnv_string,
    ((TEXTNS, u'key2'), None): cnv_string,
    ((TEXTNS, u'key2-phonetic'), None): cnv_string,
    ((TEXTNS, u'kind'), None): cnv_string,
    ((TEXTNS, u'label'), None): cnv_string,
    ((TEXTNS, u'last-row-end-column'), None): cnv_string,
    ((TEXTNS, u'last-row-start-column'), None): cnv_string,
    ((TEXTNS, u'level'), None): cnv_positiveInteger,
    ((TEXTNS, u'line-break'), None): cnv_boolean,
    ((TEXTNS, u'line-number'), None): cnv_string,
    ((TEXTNS, u'main-entry'), None): cnv_boolean,
    ((TEXTNS, u'main-entry-style-name'), None): cnv_StyleNameRef,
    ((TEXTNS, u'master-page-name'), None): cnv_StyleNameRef,
    ((TEXTNS, u'min-label-distance'), None): cnv_string,
    ((TEXTNS, u'min-label-width'), None): cnv_string,
    ((TEXTNS, u'month'), None): cnv_string,
    ((TEXTNS, u'name'), None): cnv_string,
    ((TEXTNS, u'note-class'), None): cnv_textnoteclass,
    ((TEXTNS, u'note'), None): cnv_string,
    ((TEXTNS, u'number'), None): cnv_string,
    ((TEXTNS, u'number-lines'), None): cnv_boolean,
    ((TEXTNS, u'number-position'), None): cnv_string,
    ((TEXTNS, u'numbered-entries'), None): cnv_boolean,
    ((TEXTNS, u'offset'), None): cnv_string,
    ((TEXTNS, u'organizations'), None): cnv_string,
    ((TEXTNS, u'outline-level'), None): cnv_string,
    ((TEXTNS, u'page-adjust'), None): cnv_integer,
    ((TEXTNS, u'pages'), None): cnv_string,
    ((TEXTNS, u'paragraph-style-name'), None): cnv_StyleNameRef,
    ((TEXTNS, u'placeholder-type'), None): cnv_string,
    ((TEXTNS, u'prefix'), None): cnv_string,
    ((TEXTNS, u'protected'), None): cnv_boolean,
    ((TEXTNS, u'protection-key'), None): cnv_string,
    ((TEXTNS, u'publisher'), None): cnv_string,
    ((TEXTNS, u'ref-name'), None): cnv_string,
    ((TEXTNS, u'reference-format'), None): cnv_string,
    ((TEXTNS, u'relative-tab-stop-position'), None): cnv_boolean,
    ((TEXTNS, u'report-type'), None): cnv_string,
    ((TEXTNS, u'restart-numbering'), None): cnv_boolean,
    ((TEXTNS, u'restart-on-page'), None): cnv_boolean,
    ((TEXTNS, u'row-number'), None): cnv_nonNegativeInteger,
    ((TEXTNS, u'school'), None): cnv_string,
    ((TEXTNS, u'section-name'), None): cnv_string,
    ((TEXTNS, u'select-page'), None): cnv_string,
    ((TEXTNS, u'separation-character'), None): cnv_string,
    ((TEXTNS, u'series'), None): cnv_string,
    ((TEXTNS, u'sort-algorithm'), None): cnv_string,
    ((TEXTNS, u'sort-ascending'), None): cnv_boolean,
    ((TEXTNS, u'sort-by-position'), None): cnv_boolean,
    ((TEXTNS, u'space-before'), None): cnv_string,
    ((TEXTNS, u'start-numbering-at'), None): cnv_string,
    ((TEXTNS, u'start-value'), None): cnv_nonNegativeInteger,
    ((TEXTNS, u'start-value'), None): cnv_positiveInteger,
    ((TEXTNS, u'string-value'), None): cnv_string,
    ((TEXTNS, u'string-value-if-false'), None): cnv_string,
    ((TEXTNS, u'string-value-if-true'), None): cnv_string,
    ((TEXTNS, u'string-value-phonetic'), None): cnv_string,
    ((TEXTNS, u'style-name'), None): cnv_StyleNameRef,
    ((TEXTNS, u'suffix'), None): cnv_string,
    ((TEXTNS, u'tab-ref'), None): cnv_nonNegativeInteger,
    ((TEXTNS, u'table-name'), None): cnv_string,
    ((TEXTNS, u'table-type'), None): cnv_string,
    ((TEXTNS, u'time-adjust'), None): cnv_duration,
    ((TEXTNS, u'time-value'), None): cnv_dateTime,
    ((TEXTNS, u'time-value'), None): cnv_time,
    ((TEXTNS, u'title'), None): cnv_string,
    ((TEXTNS, u'track-changes'), None): cnv_boolean,
    ((TEXTNS, u'url'), None): cnv_string,
    ((TEXTNS, u'use-caption'), None): cnv_boolean,
    ((TEXTNS, u'use-chart-objects'), None): cnv_boolean,
    ((TEXTNS, u'use-draw-objects'), None): cnv_boolean,
    ((TEXTNS, u'use-floating-frames'), None): cnv_boolean,
    ((TEXTNS, u'use-graphics'), None): cnv_boolean,
    ((TEXTNS, u'use-index-marks'), None): cnv_boolean,
    ((TEXTNS, u'use-index-source-styles'), None): cnv_boolean,
    ((TEXTNS, u'use-keys-as-entries'), None): cnv_boolean,
    ((TEXTNS, u'use-math-objects'), None): cnv_boolean,
    ((TEXTNS, u'use-objects'), None): cnv_boolean,
    ((TEXTNS, u'use-other-objects'), None): cnv_boolean,
    ((TEXTNS, u'use-outline-level'), None): cnv_boolean,
    ((TEXTNS, u'use-soft-page-breaks'), None): cnv_boolean,
    ((TEXTNS, u'use-spreadsheet-objects'), None): cnv_boolean,
    ((TEXTNS, u'use-tables'), None): cnv_boolean,
    ((TEXTNS, u'value'), None): cnv_nonNegativeInteger,
    ((TEXTNS, u'visited-style-name'), None): cnv_StyleNameRef,
    ((TEXTNS, u'volume'), None): cnv_string,
    ((TEXTNS, u'year'), None): cnv_string,
    ((XFORMSNS, u'bind'), None): cnv_string,
    ((XLINKNS, u'actuate'), None): cnv_string,
    ((XLINKNS, u'href'), None): cnv_anyURI,
    ((XLINKNS, u'show'), None): cnv_xlinkshow,
    ((XLINKNS, u'title'), None): cnv_string,
    ((XLINKNS, u'type'), None): cnv_string,
}


class AttrConverters:
    def convert(self, attribute, value, element):
        """ Based on the element, figures out how to check/convert the attribute value
            All values are converted to string
        """
        conversion = attrconverters.get((attribute, element.qname), None)
        if conversion is not None:
            return conversion(attribute, value, element)
        else:
            conversion = attrconverters.get((attribute, None), None)
            if conversion is not None:
                return conversion(attribute, value, element)
        return unicode(value)



# Copyright (c) 2009-11, Walter Bender

# This procedure is invoked when the user-definable block on the
# "extras" palette is selected.

# Usage: Import this code into a Python (user-definable) block; when
# this code is run, the current hour, minute, and second are pushed to
# the FILO heap. To use these values, pop second, then minute, then
# hour from the FILO.

from time import localtime


def myblock(tw, x):  # ignore second argument
    ''' Push hours, minutes, seconds onto the FILO. '''

    # Use three 'pop' blocks to retrieve these values.
    # Note: because we use a FILO (first in, last out heap),
    # the first value you pop off of the FILO will be seconds.

    tw.lc.heap.append(localtime().tm_hour)
    tw.lc.heap.append(localtime().tm_min)
    tw.lc.heap.append(localtime().tm_sec)



#!/usr/bin/env python
# Copyright (C) 2010,11 Emiliano Pastorino <epastorino@plan.ceibal.edu.uy>
# Copyright (c) 2011 Walter Bender
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import dbus
from dbus.mainloop.glib import DBusGMainLoop
from gettext import gettext as _

from plugins.rfid.rfidutils import strhex2bin, strbin2dec, find_device
from plugins.plugin import Plugin

from TurtleArt.tapalette import make_palette
from TurtleArt.tautils import debug_output
from TurtleArt.taprimitive import Primitive
from TurtleArt.tatype import TYPE_STRING

import logging
_logger = logging.getLogger('turtleart-activity RFID plugin')

HAL_SERVICE = 'org.freedesktop.Hal'
HAL_MGR_PATH = '/org/freedesktop/Hal/Manager'
HAL_MGR_IFACE = 'org.freedesktop.Hal.Manager'
HAL_DEV_IFACE = 'org.freedesktop.Hal.Device'
REGEXP_SERUSB = '\/org\/freedesktop\/Hal\/devices\/usb_device['\
                'a-z,A-Z,0-9,_]*serial_usb_[0-9]'


class Rfid(Plugin):

    def __init__(self, parent):
        Plugin.__init__(self)
        self._parent = parent
        self._status = False

        """
        The following code will initialize a USB RFID reader. Please note that
        in order to make this initialization function work, it is necessary to
        set the permission for the ttyUSB device to 0666. You can do this by
        adding a rule to /etc/udev/rules.d

        As root (using sudo or su), copy the following text into a new file in
        /etc/udev/rules.d/94-ttyUSB-rules

        KERNEL=="ttyUSB[0-9]",MODE="0666"

        You only have to do this once.
        """

        self.rfid_connected = False
        self.rfid_device = find_device()
        self.rfid_idn = ''

        if self.rfid_device is not None:
            _logger.info("RFID device found")
            self.rfid_connected = self.rfid_device.do_connect()
            if self.rfid_connected:
                self.rfid_device.connect("tag-read", self._tag_read_cb)
                self.rfid_device.connect("disconnected", self._disconnected_cb)

            loop = DBusGMainLoop()
            bus = dbus.SystemBus(mainloop=loop)
            hmgr_iface = dbus.Interface(
                bus.get_object(
                    HAL_SERVICE,
                    HAL_MGR_PATH),
                HAL_MGR_IFACE)

            hmgr_iface.connect_to_signal('DeviceAdded', self._device_added_cb)

            self._status = True

    def setup(self):
        # set up RFID-specific blocks
        palette = make_palette('sensor',
                               colors=["#FF6060", "#A06060"],
                               help_string=_('Palette of sensor blocks'),
                               position=6)

        if self._status:
            palette.add_block('rfid',
                              style='box-style',
                              label=_('RFID'),
                              help_string=_('read value from RFID device'),
                              value_block=True,
                              prim_name='rfid')
        else:
            palette.add_block('rfid',
                              hidden=True,
                              style='box-style',
                              label=_('RFID'),
                              help_string=_('read value from RFID device'),
                              value_block=True,
                              prim_name='rfid')

        self._parent.lc.def_prim(
            'rfid', 0,
            Primitive(self.prim_read_rfid,
                      return_type=TYPE_STRING,
                      call_afterwards=self.after_rfid))

    def _status_report(self):
        debug_output('Reporting RFID status: %s' % (str(self._status)))
        return self._status

    def _device_added_cb(self, path):
        """
        Called from hal connection when a new device is plugged.
        """
        if not self.rfid_connected:
            self.rfid_device = find_device()
            _logger.debug("DEVICE_ADDED: %s" % self.rfid_device)
            if self.rfid_device is not None:
                _logger.debug("DEVICE_ADDED: RFID device is not None!")
                self.rfid_connected = self._device.do_connect()
            if self.rfid_connected:
                _logger.debug("DEVICE_ADDED: Connected!")
                self.rfid_device.connect("tag-read", self._tag_read_cb)
                self.rfid_device.connect("disconnected", self._disconnected_cb)

    def _disconnected_cb(self, device, text):
        """
        Called when the device is disconnected.
        """
        self.rfid_connected = False
        self.rfid_device = None

    def _tag_read_cb(self, device, tagid):
        """
        Callback for "tag-read" signal. Receives the read tag id.
        """
        idbin = strhex2bin(tagid)
        self.rfid_idn = strbin2dec(idbin[26:64])
        while self.rfid_idn.__len__() < 9:
            self.rfid_idn = '0' + self.rfid_idn
        print tagid, idbin, self.rfid_idn
        self.tw.lc.update_label_value('rfid', self.rfid_idn)

    # Block primitives used in talogo

    def prim_read_rfid(self):
        if self._status:
            return self.rfid_idn
        else:
            return '0'

    def after_rfid(self):
        if self._parent.lc.update_values:
            self._parent.lc.update_label_value('rfid', self.rfid_idn)
